# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path
import io
import json
import pytest
import shutil
import subprocess

from glean_parser import javascript_server
from glean_parser import translate
from glean_parser import validate_ping
from glean_parser.metrics import Metric
from unittest.mock import Mock


ROOT = Path(__file__).parent


def test_parser_js_server_ping_no_metrics(tmp_path):
    """Test that no files are generated if only ping definitions
    are provided without any metrics."""
    translate.translate(
        ROOT / "data" / "fxa-server-pings.yaml",
        "javascript_server",
        tmp_path,
    )

    assert all(False for _ in tmp_path.iterdir())


def test_parser_js_server_metrics_no_ping(tmp_path):
    """Test that no files are generated if only metric definitions
    are provided without pings."""
    translate.translate(
        ROOT / "data" / "fxa-server-metrics.yaml",
        "javascript_server",
        tmp_path,
    )

    assert all(False for _ in tmp_path.iterdir())


def test_parser_js_server(tmp_path):
    """Test that no files are generated if only metric definitions
    are provided without pings."""
    translate.translate(
        [
            ROOT / "data" / "fxa-server-pings.yaml",
            ROOT / "data" / "fxa-server-metrics.yaml",
        ],
        "javascript_server",
        tmp_path,
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["server_events.js"])

    # Make sure string metric made it in
    with (tmp_path / "server_events.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "'event.name': event_name" in content


def test_generate_ping_factory_method():
    ping = "accounts_events"
    expected_result = "createAccountsEventsEvent"
    result = javascript_server.generate_ping_factory_method(ping, metrics_by_type={})
    assert result == expected_result

    ping = "accounts_events"
    expected_result = "createAccountsEventsServerEventLogger"
    result = javascript_server.generate_ping_factory_method(
        ping, metrics_by_type={"event": [Mock(spec=Metric)]}
    )
    assert result == expected_result


def run_logger(code_dir, import_file, factory, code):
    """
    Run the JavaScript logger with a mocked logger
    that just prints the ping payload to STDOUT.
    """

    shutil.copy(ROOT / "test-js" / "package.json", code_dir)
    subprocess.check_call(["npm", "install"], cwd=code_dir)

    tmpl_code = ""
    with open(ROOT / "test-js" / "test.js.tmpl", "r") as fp:
        tmpl_code = fp.read()

    tmpl_code = (
        tmpl_code.replace("/* IMPORT */", import_file)
        .replace("/* FACTORY */", factory)
        .replace("/* CODE */", code)
    )

    with open(code_dir / "test.js", "w") as fp:
        fp.write(tmpl_code)

    return subprocess.check_output(["node", "test.js"], cwd=code_dir).decode("utf-8")


@pytest.mark.node_dependency
def test_logging_custom_ping_as_events(tmp_path):
    translate.translate(
        [
            ROOT / "data" / "fxa-server-pings.yaml",
            ROOT / "data" / "fxa-server-metrics.yaml",
        ],
        "javascript_server",
        tmp_path,
    )

    factory = "createAccountsEventsEvent"
    code = """
eventLogger.record({ user_agent: "glean-test/1.0", event_name: "testing" });
    """

    logged_output = run_logger(tmp_path, "server_events.js", factory, code)
    logged_output = json.loads(logged_output)
    fields = logged_output["Fields"]
    payload = fields["payload"]

    assert "glean-server-event" == logged_output["Type"]
    assert "glean.test" == fields["document_namespace"]
    assert "accounts-events" == fields["document_type"]
    assert "1" == fields["document_version"]
    assert "glean-test/1.0" == fields["user_agent"]

    schema_url = (
        "https://raw.githubusercontent.com/mozilla-services/"
        "mozilla-pipeline-schemas/main/"
        "schemas/glean/glean/glean.1.schema.json"
    )

    input = io.StringIO(payload)
    output = io.StringIO()
    assert (
        validate_ping.validate_ping(input, output, schema_url=schema_url) == 0
    ), output.getvalue()


@pytest.mark.node_dependency
def test_logging_events_ping_with_event_metrics(tmp_path):
    translate.translate(
        [
            ROOT / "data" / "server_metrics_with_event.yaml",
        ],
        "javascript_server",
        tmp_path,
    )

    factory = "createEventsServerEventLogger"
    code = """
eventLogger.recordBackendObjectUpdate({
  user_agent: 'glean-test/1.0',
  ip_address: '2a02:a311:803c:6300:4074:5cf2:91ac:d546',
  identifiers_fxa_account_id: 'abc',
  object_type: 'unknown',
  object_state: 'great',
  linking: true,
});
    """

    logged_output = run_logger(tmp_path, "server_events.js", factory, code)
    logged_output = json.loads(logged_output)
    fields = logged_output["Fields"]
    payload = fields["payload"]

    assert "glean-server-event" == logged_output["Type"]
    assert "glean.test" == fields["document_namespace"]
    assert "events" == fields["document_type"]
    assert "1" == fields["document_version"]
    assert "glean-test/1.0" == fields["user_agent"]

    schema_url = (
        "https://raw.githubusercontent.com/mozilla-services/"
        "mozilla-pipeline-schemas/main/"
        "schemas/glean/glean/glean.1.schema.json"
    )

    input = io.StringIO(payload)
    output = io.StringIO()
    assert (
        validate_ping.validate_ping(input, output, schema_url=schema_url) == 0
    ), output.getvalue()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import enum
from pathlib import Path
import re
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Iterable,
    Optional,
    Tuple,
    Union,
)  # noqa


from . import metrics
from . import parser
from . import pings
from . import tags
from . import util


# Yield only an error message
LintGenerator = Generator[str, None, None]

# Yield fully constructed GlinterNits
NitGenerator = Generator["GlinterNit", None, None]


class CheckType(enum.Enum):
    warning = 0
    error = 1


def _split_words(name: str) -> List[str]:
    """
    Helper function to split words on either `.` or `_`.
    """
    return re.split("[._-]", name)


def _english_list(items: List[str]) -> str:
    """
    Helper function to format a list [A, B, C] as "'A', 'B', or 'C'".
    """
    if len(items) == 0:
        return ""
    elif len(items) == 1:
        return f"'{items[0]}'"
    else:
        return "{}, or '{}'".format(
            ", ".join([f"'{x}'" for x in items[:-1]]), items[-1]
        )


def _hamming_distance(str1: str, str2: str) -> int:
    """
    Count the # of differences between strings str1 and str2,
    padding the shorter one with whitespace
    """

    diffs = 0
    if len(str1) < len(str2):
        str1, str2 = str2, str1
    len_dist = len(str1) - len(str2)
    str2 += " " * len_dist

    for ch1, ch2 in zip(str1, str2):
        if ch1 != ch2:
            diffs += 1
    return diffs


def check_common_prefix(
    category_name: str, metrics: Iterable[metrics.Metric]
) -> LintGenerator:
    """
    Check if all metrics begin with a common prefix.
    """
    metric_words = sorted([_split_words(metric.name) for metric in metrics])

    if len(metric_words) < 2:
        return

    first = metric_words[0]
    last = metric_words[-1]

    for i in range(min(len(first), len(last))):
        if first[i] != last[i]:
            break

    if i > 0:
        common_prefix = "_".join(first[:i])
        yield (
            f"Within category '{category_name}', all metrics begin with "
            f"prefix '{common_prefix}'."
            "Remove the prefixes on the metric names and (possibly) "
            "rename the category."
        )


def check_unit_in_name(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    """
    The metric name ends in a unit.
    """
    TIME_UNIT_ABBREV = {
        "nanosecond": "ns",
        "microsecond": "us",
        "millisecond": "ms",
        "second": "s",
        "minute": "m",
        "hour": "h",
        "day": "d",
    }

    MEMORY_UNIT_ABBREV = {
        "byte": "b",
        "kilobyte": "kb",
        "megabyte": "mb",
        "gigabyte": "gb",
    }

    name_words = _split_words(metric.name)
    unit_in_name = name_words[-1]

    time_unit = getattr(metric, "time_unit", None)
    memory_unit = getattr(metric, "memory_unit", None)
    unit = getattr(metric, "unit", None)

    if time_unit is not None:
        if (
            unit_in_name == TIME_UNIT_ABBREV.get(time_unit.name)
            or unit_in_name == time_unit.name
        ):
            yield (
                f"Suffix '{unit_in_name}' is redundant with time_unit "
                f"'{time_unit.name}'. Only include time_unit."
            )
        elif (
            unit_in_name in TIME_UNIT_ABBREV.keys()
            or unit_in_name in TIME_UNIT_ABBREV.values()
        ):
            yield (
                f"Suffix '{unit_in_name}' doesn't match time_unit "
                f"'{time_unit.name}'. "
                "Confirm the unit is correct and only include time_unit."
            )

    elif memory_unit is not None:
        if (
            unit_in_name == MEMORY_UNIT_ABBREV.get(memory_unit.name)
            or unit_in_name == memory_unit.name
        ):
            yield (
                f"Suffix '{unit_in_name}' is redundant with memory_unit "
                f"'{memory_unit.name}'. "
                "Only include memory_unit."
            )
        elif (
            unit_in_name in MEMORY_UNIT_ABBREV.keys()
            or unit_in_name in MEMORY_UNIT_ABBREV.values()
        ):
            yield (
                f"Suffix '{unit_in_name}' doesn't match memory_unit "
                f"{memory_unit.name}'. "
                "Confirm the unit is correct and only include memory_unit."
            )

    elif unit is not None:
        if unit_in_name == unit:
            yield (
                f"Suffix '{unit_in_name}' is redundant with unit param "
                f"'{unit}'. "
                "Only include unit."
            )


def check_category_generic(
    category_name: str, metrics: Iterable[metrics.Metric]
) -> LintGenerator:
    """
    The category name is too generic.
    """
    GENERIC_CATEGORIES = ["metrics", "events"]

    if category_name in GENERIC_CATEGORIES:
        yield (
            f"Category '{category_name}' is too generic. "
            f"Don't use {_english_list(GENERIC_CATEGORIES)} for category names"
        )


def check_bug_number(
    metric: Union[metrics.Metric, pings.Ping], parser_config: Dict[str, Any]
) -> LintGenerator:
    number_bugs = [str(bug) for bug in metric.bugs if isinstance(bug, int)]

    if len(number_bugs):
        yield (
            f"For bugs {', '.join(number_bugs)}: "
            "Bug numbers are deprecated and should be changed to full URLs. "
            f"For example, use 'http://bugzilla.mozilla.org/{number_bugs[0]}' "
            f"instead of '{number_bugs[0]}'."
        )


def check_valid_in_baseline(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    allow_reserved = parser_config.get("allow_reserved", False)

    if not allow_reserved and "baseline" in metric.send_in_pings:
        yield (
            "The baseline ping is Glean-internal. "
            "Remove 'baseline' from the send_in_pings array."
        )


def check_misspelled_pings(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    for ping in metric.send_in_pings:
        for builtin in pings.RESERVED_PING_NAMES:
            distance = _hamming_distance(ping, builtin)
            if distance == 1:
                yield f"Ping '{ping}' seems misspelled. Did you mean '{builtin}'?"


def check_tags_required(
    metric_or_ping: Union[metrics.Metric, pings.Ping], parser_config: Dict[str, Any]
) -> LintGenerator:
    if parser_config.get("require_tags", False) and not len(
        metric_or_ping.metadata.get("tags", [])
    ):
        yield "Tags are required but no tags specified"


def check_user_lifetime_expiration(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    if metric.lifetime == metrics.Lifetime.user and metric.expires != "never":
        yield (
            "Metrics with 'user' lifetime cannot have an expiration date. "
            "They live as long as the user profile does. "
            "Set expires to 'never'."
        )


def check_expired_date(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    try:
        metric.validate_expires()
    except ValueError as e:
        yield (str(e))


def check_expired_metric(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    if metric.is_expired():
        yield ("Metric has expired. Please consider removing it.")


def check_old_event_api(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    # Glean v52.0.0 removed the old events API.
    # The metrics-2-0-0 schema still supports it.
    # We want to warn about it.
    # This can go when we introduce 3-0-0

    if not isinstance(metric, metrics.Event):
        return

    if not all("type" in x for x in metric.extra_keys.values()):
        yield ("The old event API is gone. Extra keys require a type.")


def check_metric_on_events_lifetime(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    """A non-event metric on the Events ping only makes sense if its value
    is immutable over the life of the ping."""
    if (
        "events" in metric.send_in_pings
        and "all_pings" not in metric.send_in_pings
        and metric.type != "event"
        and metric.lifetime == metrics.Lifetime.ping
    ):
        yield (
            "Non-event metrics sent on the Events ping should not have the ping"
            " lifetime."
        )


def check_unexpected_unit(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    """
    `unit` was allowed on all metrics and recently disallowed.
    We now warn about its use on all but quantity and custom distribution
    metrics.
    """
    allowed_types = [metrics.Quantity, metrics.CustomDistribution]
    if not any([isinstance(metric, ty) for ty in allowed_types]) and metric.unit:
        yield (
            "The `unit` property is only allowed for quantity "
            + "and custom distribution metrics."
        )


def check_empty_datareview(
    metric: metrics.Metric, parser_config: Dict[str, Any]
) -> LintGenerator:
    disallowed_datareview = ["", "todo"]
    data_reviews = [dr.lower() in disallowed_datareview for dr in metric.data_reviews]

    if any(data_reviews):
        yield "List of data reviews should not contain empty strings or TODO markers."


def check_redundant_ping(
    pings: pings.Ping, parser_config: Dict[str, Any]
) -> LintGenerator:
    """
    Check if the pings contains 'ping' as the prefix or suffix, or 'ping' or 'custom'
    """
    ping_words = _split_words(pings.name)

    if len(ping_words) != 0:
        ping_first_word = ping_words[0]
        ping_last_word = ping_words[-1]

        if ping_first_word == "ping":
            yield ("The prefix 'ping' is redundant.")
        elif ping_last_word == "ping":
            yield ("The suffix 'ping' is redundant.")
        elif "ping" in ping_words:
            yield ("The word 'ping' is redundant.")
        elif "custom" in ping_words:
            yield ("The word 'custom' is redundant.")


def check_unknown_ping(
    check_name: str,
    check_type: CheckType,
    all_pings: Dict[str, pings.Ping],
    metrics: Dict[str, metrics.Metric],
    parser_config: Dict[str, Any],
) -> NitGenerator:
    """
    Check that all pings in `send_in_pings` for all metrics are either a builtin ping
    or in the list of defined custom pings.
    """
    available_pings = [p for p in all_pings]

    for _, metric in metrics.items():
        if check_name in metric.no_lint:
            continue

        send_in_pings = metric.send_in_pings
        for target_ping in send_in_pings:
            if target_ping in pings.RESERVED_PING_NAMES:
                continue

            if target_ping not in available_pings:
                msg = f"Ping `{target_ping} `in `send_in_pings` is unknown."
                name = ".".join([metric.category, metric.name])
                nit = GlinterNit(
                    check_name,
                    name,
                    msg,
                    check_type,
                )
                yield nit


# The checks that operate on an entire category of metrics:
#    {NAME: (function, is_error)}
CATEGORY_CHECKS: Dict[
    str, Tuple[Callable[[str, Iterable[metrics.Metric]], LintGenerator], CheckType]
] = {
    "COMMON_PREFIX": (check_common_prefix, CheckType.error),
    "CATEGORY_GENERIC": (check_category_generic, CheckType.error),
}


# The checks that operate on individual metrics:
#     {NAME: (function, is_error)}
METRIC_CHECKS: Dict[
    str, Tuple[Callable[[metrics.Metric, dict], LintGenerator], CheckType]
] = {
    "UNIT_IN_NAME": (check_unit_in_name, CheckType.error),
    "BUG_NUMBER": (check_bug_number, CheckType.error),
    "BASELINE_PING": (check_valid_in_baseline, CheckType.error),
    "MISSPELLED_PING": (check_misspelled_pings, CheckType.error),
    "TAGS_REQUIRED": (check_tags_required, CheckType.error),
    "EXPIRATION_DATE_TOO_FAR": (check_expired_date, CheckType.warning),
    "USER_LIFETIME_EXPIRATION": (check_user_lifetime_expiration, CheckType.warning),
    "EXPIRED": (check_expired_metric, CheckType.warning),
    "OLD_EVENT_API": (check_old_event_api, CheckType.warning),
    "METRIC_ON_EVENTS_LIFETIME": (check_metric_on_events_lifetime, CheckType.error),
    "UNEXPECTED_UNIT": (check_unexpected_unit, CheckType.warning),
    "EMPTY_DATAREVIEW": (check_empty_datareview, CheckType.warning),
}


# The checks that operate on individual pings:
#     {NAME: (function, is_error)}
PING_CHECKS: Dict[
    str, Tuple[Callable[[pings.Ping, dict], LintGenerator], CheckType]
] = {
    "BUG_NUMBER": (check_bug_number, CheckType.error),
    "TAGS_REQUIRED": (check_tags_required, CheckType.error),
    "REDUNDANT_PING": (check_redundant_ping, CheckType.error),
}

ALL_OBJECT_CHECKS: Dict[
    str,
    Tuple[
        Callable[
            # check name, check type, pings, metrics, config
            [str, CheckType, dict, dict, dict],
            NitGenerator,
        ],
        CheckType,
    ],
] = {
    "UNKNOWN_PING_REFERENCED": (check_unknown_ping, CheckType.error),
}


class GlinterNit:
    def __init__(self, check_name: str, name: str, msg: str, check_type: CheckType):
        self.check_name = check_name
        self.name = name
        self.msg = msg
        self.check_type = check_type

    def format(self):
        return (
            f"{self.check_type.name.upper()}: {self.check_name}: "
            f"{self.name}: {self.msg}"
        )


def _lint_item_tags(
    item_name: str,
    item_type: str,
    item_tag_names: List[str],
    valid_tag_names: List[str],
) -> List[GlinterNit]:
    invalid_tags = [tag for tag in item_tag_names if tag not in valid_tag_names]
    return (
        [
            GlinterNit(
                "INVALID_TAGS",
                item_name,
                f"Invalid tags specified in {item_type}: {', '.join(invalid_tags)}",
                CheckType.error,
            )
        ]
        if len(invalid_tags)
        else []
    )


def _lint_pings(
    category: Dict[str, Union[metrics.Metric, pings.Ping, tags.Tag]],
    parser_config: Dict[str, Any],
    valid_tag_names: List[str],
) -> List[GlinterNit]:
    nits: List[GlinterNit] = []

    for ping_name, ping in sorted(list(category.items())):
        assert isinstance(ping, pings.Ping)
        for check_name, (check_func, check_type) in PING_CHECKS.items():
            new_nits = list(check_func(ping, parser_config))
            if len(new_nits):
                if check_name not in ping.no_lint:
                    nits.extend(
                        GlinterNit(
                            check_name,
                            ping_name,
                            msg,
                            check_type,
                        )
                        for msg in new_nits
                    )
        nits.extend(
            _lint_item_tags(
                ping_name,
                "ping",
                ping.metadata.get("tags", []),
                valid_tag_names,
            )
        )
    return nits


def _lint_all_objects(
    objects: Dict[str, Dict[str, Union[metrics.Metric, pings.Ping, tags.Tag]]],
    parser_config: Dict[str, Any],
) -> List[GlinterNit]:
    nits: List[GlinterNit] = []

    pings = objects.get("pings")
    if not pings:
        return []

    metrics = objects.get("all_metrics")
    if not metrics:
        return []

    for check_name, (check_func, check_type) in ALL_OBJECT_CHECKS.items():
        new_nits = list(
            check_func(check_name, check_type, pings, metrics, parser_config)
        )
        nits.extend(new_nits)

    return nits


def lint_metrics(
    objs: metrics.ObjectTree,
    parser_config: Optional[Dict[str, Any]] = None,
    file=sys.stderr,
) -> List[GlinterNit]:
    """
    Performs glinter checks on a set of metrics objects.

    :param objs: Tree of metric objects, as returns by `parser.parse_objects`.
    :param file: The stream to write errors to.
    :returns: List of nits.
    """
    if parser_config is None:
        parser_config = {}

    nits: List[GlinterNit] = []
    valid_tag_names = [tag for tag in objs.get("tags", [])]

    nits.extend(_lint_all_objects(objs, parser_config))

    for category_name, category in sorted(list(objs.items())):
        if category_name == "pings":
            nits.extend(_lint_pings(category, parser_config, valid_tag_names))
            continue

        if category_name == "tags":
            # currently we have no linting for tags
            continue

        # Make sure the category has only Metrics, not Pings or Tags
        category_metrics = dict(
            (name, metric)
            for (name, metric) in category.items()
            if isinstance(metric, metrics.Metric)
        )

        for cat_check_name, (cat_check_func, check_type) in CATEGORY_CHECKS.items():
            if any(
                cat_check_name in metric.no_lint for metric in category_metrics.values()
            ):
                continue
            nits.extend(
                GlinterNit(cat_check_name, category_name, msg, check_type)
                for msg in cat_check_func(category_name, category_metrics.values())
            )

        for _metric_name, metric in sorted(list(category_metrics.items())):
            for check_name, (check_func, check_type) in METRIC_CHECKS.items():
                new_nits = list(check_func(metric, parser_config))
                if len(new_nits):
                    if check_name not in metric.no_lint:
                        nits.extend(
                            GlinterNit(
                                check_name,
                                ".".join([metric.category, metric.name]),
                                msg,
                                check_type,
                            )
                            for msg in new_nits
                        )

            # also check that tags for metric are valid
            nits.extend(
                _lint_item_tags(
                    ".".join([metric.category, metric.name]),
                    "metric",
                    metric.metadata.get("tags", []),
                    valid_tag_names,
                )
            )

    if len(nits):
        print("Sorry, Glean found some glinter nits:", file=file)
        for nit in nits:
            print(nit.format(), file=file)
        print("", file=file)
        print("Please fix the above nits to continue.", file=file)
        print(
            "To disable a check, add a `no_lint` parameter "
            "with a list of check names to disable.\n"
            "This parameter can appear with each individual metric, or at the "
            "top-level to affect the entire file.",
            file=file,
        )

    return nits


def glinter(
    input_filepaths: Iterable[Path],
    parser_config: Optional[Dict[str, Any]] = None,
    file=sys.stderr,
) -> int:
    """
    Commandline helper for glinter.

    :param input_filepaths: List of Path objects to load metrics from.
    :param parser_config: Parser configuration object, passed to
      `parser.parse_objects`.
    :param file: The stream to write the errors to.
    :return: Non-zero if there were any glinter errors.
    """
    if parser_config is None:
        parser_config = {}

    errors = 0

    objs = parser.parse_objects(input_filepaths, parser_config)
    errors += util.report_validation_errors(objs)

    nits = lint_metrics(objs.value, parser_config=parser_config, file=file)
    errors += len([nit for nit in nits if nit.check_type == CheckType.error])

    if errors == 0:
        print("✨ Your metrics are Glean! ✨", file=file)
        return 0

    print(f"❌ Found {errors} errors.")

    return 1



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path

from glean_parser import markdown
from glean_parser import metrics
from glean_parser import pings
from glean_parser import translate


ROOT = Path(__file__).parent


def test_parser(tmp_path):
    """Test translating metrics to Markdown files."""
    translate.translate(
        ROOT / "data" / "core.yaml",
        "markdown",
        tmp_path,
        {"namespace": "Foo", "introduction_extra": "Extra Intro Text Bar"},
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["metrics.md"])

    # Make sure descriptions made it in
    with (tmp_path / "metrics.md").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "is assembled out of the box by the Glean SDK." in content
        # Make sure the table structure is in place
        assert (
            "| Name | Type | Description | Data reviews | Extras | "
            + "Expiration | [Data Sensitivity]"
            in content
        )
        # Make sure non ASCII characters are there
        assert "جمع 搜集" in content
        # test that extra text made it
        assert "Extra Intro Text" in content


def test_extra_info_generator():
    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
        extra_keys={"my_extra": {"description": "an extra", "type": "string"}},
    )

    assert markdown.extra_info(event) == [("my_extra", "an extra")]

    labeled = metrics.LabeledCounter(
        type="labeled_counter",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
        labels=["label"],
    )

    assert markdown.extra_info(labeled) == [("label", None)]

    # We currently support extra info only for events and labeled types.
    other = metrics.Timespan(
        type="timespan",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        time_unit="day",
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
    )

    assert len(markdown.extra_info(other)) == 0


def test_ping_desc():
    # Make sure to return something for built-in pings.
    for ping_name in pings.RESERVED_PING_NAMES:
        assert len(markdown.ping_desc(ping_name)) > 0

    # We don't expect nothing for unknown pings.
    assert len(markdown.ping_desc("unknown-ping")) == 0

    # If we have a custom ping cache, try look up the
    # description there.
    cache = {}
    cache["cached-ping"] = pings.Ping(
        name="cached-ping",
        description="the description for the custom ping\n with a surprise",
        bugs=["1234"],
        notification_emails=["email@example.com"],
        data_reviews=["https://www.example.com/review"],
        include_client_id=False,
    )

    assert (
        markdown.ping_desc("cached-ping", cache)
        == "the description for the custom ping\n with a surprise"
    )

    # We don't expect nothing for unknown pings, even with caches.
    assert len(markdown.ping_desc("unknown-ping", cache)) == 0


def test_ping_docs():
    # Make sure to return something for built-in pings.
    for ping_name in pings.RESERVED_PING_NAMES:
        docs = markdown.ping_docs(ping_name)
        assert docs.startswith("https://")
        assert len(docs) > 0

    # We don't expect nothing for unknown pings.
    assert len(markdown.ping_docs("unknown-ping")) == 0


def test_metrics_docs():
    assert (
        markdown.metrics_docs("boolean")
        == "https://mozilla.github.io/glean/book/user/metrics/boolean.html"
    )
    assert (
        markdown.metrics_docs("labeled_counter")
        == "https://mozilla.github.io/glean/book/user/metrics/labeled_counters.html"
    )
    assert (
        markdown.metrics_docs("labeled_string")
        == "https://mozilla.github.io/glean/book/user/metrics/labeled_strings.html"
    )


def test_review_title():
    index = 1

    assert (
        markdown.ping_review_title(
            "https://bugzilla.mozilla.org/show_bug.cgi?id=1581647", index
        )
        == "Bug 1581647"
    )
    assert (
        markdown.ping_review_title(
            "https://github.com/mozilla-mobile/fenix/pull/1707", index
        )
        == "mozilla-mobile/fenix#1707"
    )
    assert markdown.ping_review_title("http://example.com/reviews", index) == "Review 1"


def test_reasons(tmp_path):
    translate.translate(
        ROOT / "data" / "pings.yaml",
        "markdown",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["metrics.md"])

    # Make sure descriptions made it in
    with (tmp_path / "metrics.md").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "- `serious`: A serious reason for sending a ping." in content


def test_event_extra_keys_in_correct_order(tmp_path):
    """
    Assert that the extra keys appear in the parameter and the enumeration in
    the same order.

    https://bugzilla.mozilla.org/show_bug.cgi?id=1648768
    """
    translate.translate(
        ROOT / "data" / "event_key_ordering.yaml",
        "markdown",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["metrics.md"])

    with (tmp_path / "metrics.md").open("r", encoding="utf-8") as fd:
        content = fd.read()
        print(content)
        content = " ".join(content.split())
        assert (
            r"<ul><li>And1WithExtraCasing: four</li>"
            r"<li>alice: two</li>"
            r"<li>bob: three</li>"
            r"<li>charlie: one</li></ul>" in content
        )


def test_send_if_empty_metrics(tmp_path):
    translate.translate(
        [
            ROOT / "data" / "send_if_empty_with_metrics.yaml",
            ROOT / "data" / "pings.yaml",
        ],
        "markdown",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["metrics.md"])

    # Make sure descriptions made it in
    with (tmp_path / "metrics.md").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "Lorem ipsum dolor sit amet, consectetur adipiscing elit." in content


def test_data_sensitivity():
    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
        extra_keys={"my_extra": {"description": "an extra", "type": "string"}},
        data_sensitivity=["technical", "interaction"],
    )

    assert markdown.data_sensitivity_numbers(event.data_sensitivity) == "1, 2"

    assert markdown.data_sensitivity_numbers(None) == "unknown"



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Outputter to generate Swift code for metrics.
"""

import enum
import json
from pathlib import Path
from typing import Any, Dict, Optional, Union

from . import __version__
from . import metrics
from . import pings
from . import tags
from . import util

# An (imcomplete) list of reserved keywords in Swift.
# These will be replaced in generated code by their escaped form.
SWIFT_RESERVED_NAMES = ["internal", "typealias"]


def swift_datatypes_filter(value: util.JSONType) -> str:
    """
    A Jinja2 filter that renders Swift literals.

    Based on Python's JSONEncoder, but overrides:
      - dicts to use `[key: value]`
      - sets to use `[...]`
      - enums to use the like-named Swift enum
      - Rate objects to a CommonMetricData initializer
        (for external Denominators' Numerators lists)
    """

    class SwiftEncoder(json.JSONEncoder):
        def iterencode(self, value):
            if isinstance(value, dict):
                yield "["
                first = True
                for key, subvalue in value.items():
                    if not first:
                        yield ", "
                    yield from self.iterencode(key)
                    yield ": "
                    yield from self.iterencode(subvalue)
                    first = False
                yield "]"
            elif isinstance(value, enum.Enum):
                yield ("." + util.camelize(value.name))
            elif isinstance(value, list):
                yield "["
                first = True
                for subvalue in value:
                    if not first:
                        yield ", "
                    yield from self.iterencode(subvalue)
                    first = False
                yield "]"
            elif isinstance(value, set):
                yield "["
                first = True
                for subvalue in sorted(list(value)):
                    if not first:
                        yield ", "
                    yield from self.iterencode(subvalue)
                    first = False
                yield "]"
            elif value is None:
                yield "nil"
            elif isinstance(value, metrics.Rate):
                yield "CommonMetricData("
                first = True
                for arg_name in util.common_metric_args:
                    if hasattr(value, arg_name):
                        if not first:
                            yield ", "
                        yield f"{util.camelize(arg_name)}: "
                        yield from self.iterencode(getattr(value, arg_name))
                        first = False
                yield ")"
            else:
                yield from super().iterencode(value)

    return "".join(SwiftEncoder().iterencode(value))


def type_name(obj: Union[metrics.Metric, pings.Ping]) -> str:
    """
    Returns the Swift type to use for a given metric or ping object.
    """
    generate_enums = getattr(obj, "_generate_enums", [])
    if len(generate_enums):
        generic = None
        for member, suffix in generate_enums:
            if len(getattr(obj, member)):
                generic = util.Camelize(obj.name) + suffix
            else:
                if isinstance(obj, metrics.Event):
                    generic = "NoExtras"
                else:
                    generic = "No" + suffix

        return "{}<{}>".format(class_name(obj.type), generic)

    generate_structure = getattr(obj, "_generate_structure", [])
    if len(generate_structure):
        generic = util.Camelize(obj.name) + "Object"
        return "{}<{}>".format(class_name(obj.type), generic)

    return class_name(obj.type)


def extra_type_name(typ: str) -> str:
    """
    Returns the corresponding Swift type for event's extra key types.
    """

    if typ == "boolean":
        return "Bool"
    elif typ == "string":
        return "String"
    elif typ == "quantity":
        return "Int32"
    else:
        return "UNSUPPORTED"


def structure_type_name(typ: str) -> str:
    """
    Returns the corresponding Swift type for structure items.
    """

    if typ == "boolean":
        return "Bool"
    elif typ == "string":
        return "String"
    elif typ == "number":
        return "Int64"
    else:
        return "UNSUPPORTED"


def class_name(obj_type: str) -> str:
    """
    Returns the Swift class name for a given metric or ping type.
    """
    if obj_type == "ping":
        return "Ping"
    if obj_type.startswith("labeled_"):
        obj_type = obj_type[8:]
    return util.Camelize(obj_type) + "MetricType"


def variable_name(var: str) -> str:
    """
    Returns a valid Swift variable name, escaping keywords if necessary.
    """
    if var in SWIFT_RESERVED_NAMES:
        return "`" + var + "`"
    else:
        return var


class BuildInfo:
    def __init__(self, build_date):
        self.build_date = build_date


def generate_build_date(date: Optional[str]) -> str:
    """
    Generate the build timestamp.
    """

    ts = util.build_date(date)

    data = [
        ("year", ts.year),
        ("month", ts.month),
        ("day", ts.day),
        ("hour", ts.hour),
        ("minute", ts.minute),
        ("second", ts.second),
    ]

    # The internal DatetimeMetricType API can take a `DateComponents` object,
    # which lets us easily specify the timezone.
    components = ", ".join([f"{name}: {val}" for (name, val) in data])
    return f'DateComponents(calendar: Calendar.current, timeZone: TimeZone(abbreviation: "UTC"), {components})'  # noqa


class Category:
    """
    Data struct holding information about a metric to be used in the template.
    """

    name: str
    objs: Dict[str, Union[metrics.Metric, pings.Ping, tags.Tag]]
    contains_pings: bool


def output_swift(
    objs: metrics.ObjectTree, output_dir: Path, options: Optional[Dict[str, Any]] = None
) -> None:
    """
    Given a tree of objects, output Swift code to `output_dir`.

    :param objects: A tree of objects (metrics and pings) as returned from
        `parser.parse_objects`.
    :param output_dir: Path to an output directory to write to.
    :param options: options dictionary, with the following optional keys:
        - namespace: The namespace to generate metrics in
        - glean_namespace: The namespace to import Glean from
        - allow_reserved: When True, this is a Glean-internal build
        - with_buildinfo: If "true" the `GleanBuildInfo` is generated.
          Otherwise generation of that file is skipped.
          Defaults to "true".
        - build_date: If set to `0` a static unix epoch time will be used.
                      If set to a ISO8601 datetime string (e.g. `2022-01-03T17:30:00`)
                      it will use that date.
                      Other values will throw an error.
                      If not set it will use the current date & time.
    """
    if options is None:
        options = {}

    template = util.get_jinja2_template(
        "swift.jinja2",
        filters=(
            ("swift", swift_datatypes_filter),
            ("type_name", type_name),
            ("class_name", class_name),
            ("variable_name", variable_name),
            ("extra_type_name", extra_type_name),
            ("structure_type_name", structure_type_name),
        ),
    )

    namespace = options.get("namespace", "GleanMetrics")
    glean_namespace = options.get("glean_namespace", "Glean")
    with_buildinfo = options.get("with_buildinfo", "true").lower() == "true"
    build_date = options.get("build_date", None)
    build_info = None
    if with_buildinfo:
        build_date = generate_build_date(build_date)
        build_info = BuildInfo(build_date=build_date)

    filename = "Metrics.swift"
    filepath = output_dir / filename
    categories = []

    for category_key, category_val in objs.items():
        contains_pings = any(
            isinstance(obj, pings.Ping) for obj in category_val.values()
        )

        cat = Category()
        cat.name = category_key
        cat.objs = category_val
        cat.contains_pings = contains_pings

        categories.append(cat)

    with filepath.open("w", encoding="utf-8") as fd:
        fd.write(
            template.render(
                parser_version=__version__,
                categories=categories,
                common_metric_args=util.common_metric_args,
                extra_metric_args=util.extra_metric_args,
                namespace=namespace,
                glean_namespace=glean_namespace,
                allow_reserved=options.get("allow_reserved", False),
                build_info=build_info,
            )
        )
        # Jinja2 squashes the final newline, so we explicitly add it
        fd.write("\n")



import pydoc


def translate_options(ctx, param, value):
    text = """Target language options for Translate function

These are backend specific and optional, provide as key:value

Rust: no options.

Swift:
- `namespace`: The namespace to generate metrics in
- `glean_namespace`: The namespace to import Glean from
- `allow_reserved`: When True, this is a Glean-internal build
- `with_buildinfo`: If "true" the `GleanBuildInfo` is generated.
        Otherwise generation of that file is skipped.
        Defaults to "true".
- `build_date`: If set to `0` a static unix epoch time will be used.
        If set to a ISO8601 datetime string (e.g. `2022-01-03T17:30:00`)
        it will use that date.
        Other values will throw an error.
        If not set it will use the current date & time.

Kotlin:
- `namespace`: The package namespace to declare at the top of the
        generated files. Defaults to `GleanMetrics`.
- `glean_namespace`: The package namespace of the glean library itself.
        This is where glean objects will be imported from in the generated
        code.

JavaScript:
- `platform`: Which platform are we building for. Options are `webext` and `qt`.
        Default is `webext`.
- `version`: The version of the Glean.js Qt library being used.
        This option is mandatory when targeting Qt. Note that the version
        string must only contain the major and minor version i.e. 0.14.
- `with_buildinfo`: If "true" a `gleanBuildInfo.(js|ts)` file is generated.
        Otherwise generation of that file is skipped. Defaults to "false".
- `build_date`: If set to `0` a static unix epoch time will be used.
        If set to a ISO8601 datetime string (e.g. `2022-01-03T17:30:00`)
        it will use that date.
        Other values will throw an error.
        If not set it will use the current date & time.

Markdown:
- `project_title`: The project's title.

(press q to exit)"""

    if value:
        if value[0].lower() == "help":
            pydoc.pager(text)
            ctx.exit()
    return value



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path

from glean_parser import javascript
from glean_parser import metrics
from glean_parser import pings
from glean_parser import translate


ROOT = Path(__file__).parent


def test_parser_js(tmp_path):
    """Test translating metrics to Javascript files."""
    translate.translate(
        ROOT / "data" / "core.yaml",
        "javascript",
        tmp_path,
        None,
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(
        [
            "corePing.js",
            "telemetry.js",
            "environment.js",
            "dottedCategory.js",
            "gleanInternalMetrics.js",
        ]
    )

    # Make sure descriptions made it in
    with (tmp_path / "corePing.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "True if the user has set Firefox as the default browser." in content

    with (tmp_path / "telemetry.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "جمع 搜集" in content

    with (tmp_path / "gleanInternalMetrics.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert 'category: ""' in content


def test_parser_js_all_metrics(tmp_path):
    """Test translating metrics to Javascript files."""
    translate.translate(
        ROOT / "data" / "all_metrics.yaml",
        "javascript",
        tmp_path,
        None,
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["allMetrics.js"])


def test_parser_ts(tmp_path):
    """Test translating metrics to Typescript files."""
    translate.translate(
        ROOT / "data" / "core.yaml",
        "typescript",
        tmp_path,
        None,
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(
        [
            "corePing.ts",
            "telemetry.ts",
            "environment.ts",
            "dottedCategory.ts",
            "gleanInternalMetrics.ts",
        ]
    )

    # Make sure descriptions made it in
    with (tmp_path / "corePing.ts").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "use strict" not in content
        assert "True if the user has set Firefox as the default browser." in content

    with (tmp_path / "telemetry.ts").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "use strict" not in content
        assert "جمع 搜集" in content

    with (tmp_path / "gleanInternalMetrics.ts").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "use strict" not in content
        assert 'category: ""' in content


def test_ping_parser(tmp_path):
    """Test translating pings to Javascript files."""
    translate.translate(
        ROOT / "data" / "pings.yaml",
        "javascript",
        tmp_path,
        None,
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["pings.js"])

    # Make sure descriptions made it in
    with (tmp_path / "pings.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "This is a custom ping" in content


def test_javascript_generator():
    jdf = javascript.javascript_datatypes_filter
    assert jdf(metrics.Lifetime.ping) == '"ping"'


def test_metric_class_name():
    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
        extra_keys={"my_extra": {"description": "an extra", "type": "string"}},
    )

    webext_class_name = javascript.class_name_factory("webext")
    qt_class_name = javascript.class_name_factory("qt")

    assert webext_class_name(event.type) == "EventMetricType"
    assert qt_class_name(event.type) == "Glean.Glean._private.EventMetricType"

    boolean = metrics.Boolean(
        type="boolean",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
    )
    assert webext_class_name(boolean.type) == "BooleanMetricType"
    assert qt_class_name(boolean.type) == "Glean.Glean._private.BooleanMetricType"

    ping = pings.Ping(
        name="custom",
        description="description...",
        include_client_id=True,
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@nowhere.com"],
    )
    assert webext_class_name(ping.type) == "PingType"
    assert qt_class_name(ping.type) == "Glean.Glean._private.PingType"


def test_import_path():
    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
        extra_keys={"my_extra": {"description": "an extra", "type": "string"}},
    )

    assert javascript.import_path(event.type) == "metrics/event"

    boolean = metrics.Boolean(
        type="boolean",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
    )
    assert javascript.import_path(boolean.type) == "metrics/boolean"

    ping = pings.Ping(
        name="custom",
        description="description...",
        include_client_id=True,
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@nowhere.com"],
    )
    assert javascript.import_path(ping.type) == "ping"


def test_labeled_subtype_is_imported(tmp_path):
    """
    Test that both the LabeledMetricType and its subtype are imported
    """
    translate.translate(
        ROOT / "data" / "single_labeled.yaml", "javascript", tmp_path, None
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["category.js"])

    with (tmp_path / "category.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert (
            content.count(
                'import CounterMetricType from "@mozilla/glean/private/metrics/counter";'  # noqa
            )
            == 1
        )
        assert (
            content.count(
                'import LabeledMetricType from "@mozilla/glean/private/metrics/labeled";'  # noqa
            )
            == 1
        )


def test_duplicate(tmp_path):
    """
    Test that there aren't duplicate imports when using a labeled and
    non-labeled version of the same metric.

    https://github.com/mozilla-mobile/android-components/issues/2793
    """
    translate.translate(
        ROOT / "data" / "duplicate_labeled.yaml", "javascript", tmp_path, None
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["category.js"])

    with (tmp_path / "category.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert (
            content.count(
                'import CounterMetricType from "@mozilla/glean/private/metrics/counter";'  # noqa
            )
            == 1
        )


def test_reasons(tmp_path):
    translate.translate(ROOT / "data" / "pings.yaml", "javascript", tmp_path, None)

    translate.translate(ROOT / "data" / "pings.yaml", "typescript", tmp_path, None)

    assert set(x.name for x in tmp_path.iterdir()) == set(["pings.js", "pings.ts"])

    with (tmp_path / "pings.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "export const CustomPingMightBeEmptyReasonCodes" in content
        assert "export const RealPingMightBeEmptyReasonCodes" not in content

    with (tmp_path / "pings.ts").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "export enum CustomPingMightBeEmptyReasonCodes" in content
        assert "export enum RealPingMightBeEmptyReasonCodes" not in content


def test_event_extra_keys_in_correct_order(tmp_path):
    """
    Assert that the extra keys appear in the parameter and the enumeration in
    the same order.

    https://bugzilla.mozilla.org/show_bug.cgi?id=1648768
    """
    translate.translate(
        ROOT / "data" / "event_key_ordering.yaml",
        "javascript",
        tmp_path,
        None,
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["event.js"])

    with (tmp_path / "event.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        assert '["And1WithExtraCasing", "alice", "bob", "charlie"]' in content


def test_arguments_are_generated_in_deterministic_order(tmp_path):
    """
    Assert that arguments on generated code are always in the same order.

    https://bugzilla.mozilla.org/show_bug.cgi?id=1666192
    """
    translate.translate(
        ROOT / "data" / "event_key_ordering.yaml",
        "javascript",
        tmp_path,
        None,
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["event.js"])

    with (tmp_path / "event.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        expected = 'export const example = new EventMetricType({ category: "event", name: "example", sendInPings: ["events"], lifetime: "ping", disabled: false, }, ["And1WithExtraCasing", "alice", "bob", "charlie"]);'  # noqa
        assert expected in content


def test_qt_platform_template_includes_expected_imports(tmp_path):
    """
    Assert that when the platform is Qt, the template does not contain
    import/export statements.
    """
    translate.translate(
        ROOT / "data" / "single_labeled.yaml",
        "javascript",
        tmp_path,
        {"platform": "qt", "version": "0.14"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["category.js", "qmldir"])

    with (tmp_path / "category.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert content.count(".import org.mozilla.Glean 0.14") == 1
        assert content.count("export") == 0


def test_qt_platform_generated_correct_qmldir_file(tmp_path):
    """
    Assert that when the platform is Qt, a qmldir is also generated
    with the expected files listed in it.
    """
    translate.translate(
        ROOT / "data" / "core.yaml",
        "javascript",
        tmp_path,
        {"platform": "qt", "version": "0.14"},
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(
        [
            "corePing.js",
            "telemetry.js",
            "environment.js",
            "dottedCategory.js",
            "gleanInternalMetrics.js",
            "qmldir",
        ]
    )

    with (tmp_path / "qmldir").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert content.count("CorePing 0.14 corePing.js") == 1
        assert content.count("Telemetry 0.14 telemetry.js") == 1
        assert content.count("Environment 0.14 environment.js") == 1
        assert content.count("DottedCategory 0.14 dottedCategory.js") == 1
        assert content.count("GleanInternalMetrics 0.14 gleanInternalMetrics.js") == 1
        assert content.count("depends org.mozilla.Glean 0.14") == 1


def test_event_extra_keys_with_types(tmp_path):
    """
    Assert that the extra keys with types appear with their corresponding types.
    """
    translate.translate(
        ROOT / "data" / "events_with_types.yaml",
        "typescript",
        tmp_path,
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["core.ts"])

    with (tmp_path / "core.ts").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        assert (
            "new EventMetricType<{ "
            "enabled?: boolean, "
            "preference?: string, "
            "swapped?: number, "
            "}>({" in content
        )
        assert '"enabled", "preference", "swapped"' in content

    # Make sure this only happens for the TypeScript template.
    translate.translate(
        ROOT / "data" / "events_with_types.yaml",
        "javascript",
        tmp_path,
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["core.js", "core.ts"])

    with (tmp_path / "core.js").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        assert "new EventMetricType({" in content
        assert '"enabled", "preference", "swapped"' in content


def test_build_info_is_generated_when_option_is_present(tmp_path):
    """
    Assert that build info is generated
    """
    translate.translate(
        ROOT / "data" / "single_labeled.yaml",
        "typescript",
        tmp_path,
        {"with_buildinfo": "true"},
    )
    assert set(x.name for x in tmp_path.iterdir()) == set(
        ["gleanBuildInfo.ts", "category.ts"]
    )

    translate.translate(
        ROOT / "data" / "single_labeled.yaml",
        "typescript",
        tmp_path,
        {"with_buildinfo": "true", "build_date": "2022-03-01T14:10+01:00"},
    )
    assert set(x.name for x in tmp_path.iterdir()) == set(
        ["gleanBuildInfo.ts", "category.ts"]
    )
    with (tmp_path / "gleanBuildInfo.ts").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "new Date(2022, 2, 1, 14, 10, 0)" in content



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path

import pytest
import shutil

from glean_parser import parser
from glean_parser import translate
from glean_parser.util import load_yaml_or_json

import util


ROOT = Path(__file__).parent


def test_translate_unknown_format():
    with pytest.raises(ValueError) as e:
        translate.translate([], "foo", ".")

    assert "Unknown output format" in str(e)


def test_all_metrics(tmp_path):
    output = tmp_path / "tags_test"

    translate.translate(
        [ROOT / "data" / "all_metrics.yaml"],
        "kotlin",
        output,
        options={"with_buildinfo": "false"},
        parser_config={"allow_reserved": True},
    )

    # A single metric category in a single file
    assert len(list(output.iterdir())) == 1


def test_translate_missing_directory(tmp_path):
    output = tmp_path / "foo"

    translate.translate(
        ROOT / "data" / "core.yaml",
        "kotlin",
        output,
        parser_config={"allow_reserved": True},
    )

    assert len(list(output.iterdir())) == 6


def test_translate_missing_input_files(tmp_path):
    output = tmp_path

    with pytest.raises(FileNotFoundError):
        translate.translate(
            ROOT / "data" / "missing.yaml",
            "kotlin",
            output,
            parser_config={"allow_reserved": True},
        )

    assert 0 == translate.translate(
        ROOT / "data" / "missing.yaml",
        "kotlin",
        output,
        parser_config={"allow_reserved": True, "allow_missing_files": True},
    )


def test_translate_remove_obsolete_kotlin_files(tmp_path):
    output = tmp_path / "foo"

    translate.translate(
        ROOT / "data" / "core.yaml",
        "kotlin",
        output,
        parser_config={"allow_reserved": True},
    )

    assert len(list(output.iterdir())) == 6

    translate.translate(ROOT / "data" / "smaller.yaml", "kotlin", output)

    assert len(list(output.iterdir())) == 2


def test_translate_retains_existing_markdown_files(tmp_path):
    output = tmp_path / "foo"

    translate.translate(
        ROOT / "data" / "core.yaml",
        "markdown",
        output,
        parser_config={"allow_reserved": True},
    )

    # Move the file to a different location, translate always writes to
    # metrics.md.
    shutil.move(str(output / "metrics.md"), str(output / "old.md"))

    assert len(list(output.iterdir())) == 1

    translate.translate(ROOT / "data" / "smaller.yaml", "markdown", output)

    assert len(list(output.iterdir())) == 2


def test_translate_expires():
    contents = [
        {
            "metrics": {
                "a": {"type": "counter", "expires": "never"},
                "b": {"type": "counter", "expires": "expired"},
                "c": {"type": "counter", "expires": "2000-01-01"},
                "d": {"type": "counter", "expires": "2100-01-01"},
            }
        }
    ]
    contents = [util.add_required(x) for x in contents]

    objs = parser.parse_objects(contents)
    assert len(list(objs)) == 0
    objs = objs.value

    assert objs["metrics"]["a"].disabled is False
    assert objs["metrics"]["b"].disabled is True
    assert objs["metrics"]["c"].disabled is True
    assert objs["metrics"]["d"].disabled is False


def test_translate_send_in_pings(tmp_path):
    contents = [
        {
            "baseline": {
                "counter": {"type": "counter"},
                "event": {"type": "event"},
                "c": {"type": "counter", "send_in_pings": ["default", "custom"]},
            }
        }
    ]
    contents = [util.add_required(x) for x in contents]

    objs = parser.parse_objects(contents)
    assert len(list(objs)) == 0
    objs = objs.value

    assert objs["baseline"]["counter"].send_in_pings == ["metrics"]
    assert objs["baseline"]["event"].send_in_pings == ["events"]
    assert objs["baseline"]["c"].send_in_pings == ["custom", "metrics"]


def test_translate_dont_remove_extra_files(tmp_path):
    output = tmp_path / "foo"
    output.mkdir()

    with (output / "extra.txt").open("w") as fd:
        fd.write("\n")

    translate.translate(
        ROOT / "data" / "core.yaml",
        "kotlin",
        output,
        parser_config={"allow_reserved": True},
    )

    assert len(list(output.iterdir())) == 7
    assert "extra.txt" in [str(x.name) for x in output.iterdir()]


def test_external_translator(tmp_path):
    def external_translator(all_objects, output_dir, options):
        assert {"foo": "bar", "allow_reserved": True} == options

        for category in all_objects:
            with (tmp_path / category).open("w") as fd:
                for metric in category:
                    fd.write(f"{metric}\n")

    translate.translate_metrics(
        ROOT / "data" / "core.yaml",
        tmp_path,
        external_translator,
        [],
        options={"foo": "bar"},
        parser_config={"allow_reserved": True},
    )

    content = load_yaml_or_json(ROOT / "data" / "core.yaml")

    expected_keys = set(content.keys()) - set(["$schema"])

    assert set(x.name for x in tmp_path.iterdir()) == expected_keys


def test_getting_line_number():
    pings = load_yaml_or_json(ROOT / "data" / "pings.yaml")
    metrics = load_yaml_or_json(ROOT / "data" / "core.yaml")

    assert pings["custom-ping"].defined_in["line"] == 7
    assert metrics["core_ping"]["seq"].defined_in["line"] == 69


def test_rates(tmp_path):
    def external_translator(all_objects, output_dir, options):
        category = all_objects["testing.rates"]
        assert category["has_internal_denominator"].type == "rate"
        assert category["has_external_denominator"].type == "numerator"
        assert (
            category["has_external_denominator"].denominator_metric
            == "testing.rates.the_denominator"
        )
        assert category["also_has_external_denominator"].type == "numerator"
        assert (
            category["also_has_external_denominator"].denominator_metric
            == "testing.rates.the_denominator"
        )
        assert category["the_denominator"].type == "denominator"

        numerators = [
            f"{m.category}.{m.name}" for m in category["the_denominator"].numerators
        ]
        # Because generation gaurantees order, this must be sorted.
        assert numerators == sorted(
            [
                "testing.rates.has_external_denominator",
                "testing.rates.also_has_external_denominator",
            ]
        )

    translate.translate_metrics(
        ROOT / "data" / "rate.yaml",
        tmp_path,
        external_translator,
        [],
        options={"foo": "bar"},
        parser_config={"allow_reserved": True},
    )


def test_with_tags(tmp_path):
    output = tmp_path / "tags_test"

    translate.translate(
        [
            ROOT / "data" / "metric-with-tags.yaml",
            ROOT / "data" / "tags.yaml",
        ],
        "kotlin",
        output,
        options={"with_buildinfo": "false"},
        parser_config={"allow_reserved": True},
    )

    # A single metric category in a single file
    assert len(list(output.iterdir())) == 1



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Top-level package for Glean parser."""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    # package is not installed
    pass

__author__ = """The Glean Team"""
__email__ = "glean-team@mozilla.com"



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--run-web-tests",
        action="store_true",
        default=False,
        help="Run tests that require a web connection",
    )
    parser.addoption(
        "--run-node-tests",
        action="store_true",
        default=False,
        help="Run tests that require node.js",
    )
    parser.addoption(
        "--run-ruby-tests",
        action="store_true",
        default=False,
        help="Run tests that require Ruby",
    )
    parser.addoption(
        "--run-go-tests",
        action="store_true",
        default=False,
        help="Run tests that require Go",
    )


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--run-web-tests"):
        skip_web = pytest.mark.skip(reason="Need --run-web-tests option to run")
        for item in items:
            if "web_dependency" in item.keywords:
                item.add_marker(skip_web)

    if not config.getoption("--run-node-tests"):
        skip_node = pytest.mark.skip(reason="Need --run-node-tests option to run")
        for item in items:
            if "node_dependency" in item.keywords:
                item.add_marker(skip_node)

    if not config.getoption("--run-ruby-tests"):
        skip_ruby = pytest.mark.skip(reason="Need --run-ruby-tests option to run")
        for item in items:
            if "ruby_dependency" in item.keywords:
                item.add_marker(skip_ruby)

    if not config.getoption("--run-go-tests"):
        skip_go = pytest.mark.skip(reason="Need --run-go-tests option to run")
        for item in items:
            if "go_dependency" in item.keywords:
                item.add_marker(skip_go)



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Classes for managing the description of pings.
"""

from typing import Dict, List, Optional


from . import util


RESERVED_PING_NAMES = ["baseline", "metrics", "events", "deletion-request", "default"]


class Ping:
    def __init__(
        self,
        name: str,
        description: str,
        bugs: List[str],
        notification_emails: List[str],
        metadata: Optional[Dict] = None,
        data_reviews: Optional[List[str]] = None,
        include_client_id: bool = False,
        send_if_empty: bool = False,
        reasons: Optional[Dict[str, str]] = None,
        defined_in: Optional[Dict] = None,
        no_lint: Optional[List[str]] = None,
        enabled: Optional[bool] = None,
        _validated: bool = False,
    ):
        # Avoid cyclical import
        from . import parser

        self.name = name
        self.description = description

        self.bugs = bugs
        self.notification_emails = notification_emails
        if metadata is None:
            metadata = {}
        self.metadata = metadata
        self.precise_timestamps = self.metadata.get("precise_timestamps", True)
        self.include_info_sections = self.metadata.get("include_info_sections", True)
        if enabled is None:
            enabled = True
        self.enabled = enabled
        self.schedules_pings: List[str] = []
        if data_reviews is None:
            data_reviews = []
        self.data_reviews = data_reviews
        self.include_client_id = include_client_id
        self.send_if_empty = send_if_empty
        if reasons is None:
            reasons = {}
        self.reasons = reasons
        self.defined_in = defined_in
        if no_lint is None:
            no_lint = []
        self.no_lint = no_lint

        # _validated indicates whether this ping has already been jsonschema
        # validated (but not any of the Python-level validation).
        if not _validated:
            data: Dict[str, util.JSONType] = {
                "$schema": parser.PINGS_ID,
                self.name: self._serialize_input(),
            }
            for error in parser.validate(data):
                raise ValueError(error)

    _generate_enums = [("reason_codes", "ReasonCodes")]

    @property
    def type(self) -> str:
        return "ping"

    @property
    def reason_codes(self) -> List[str]:
        return sorted(list(self.reasons.keys()))

    def serialize(self) -> Dict[str, util.JSONType]:
        """
        Serialize the metric back to JSON object model.
        """
        d = self.__dict__.copy()
        del d["name"]
        return d

    def _serialize_input(self) -> Dict[str, util.JSONType]:
        d = self.serialize()
        modified_dict = util.remove_output_params(d, "defined_in")
        modified_dict = util.remove_output_params(modified_dict, "precise_timestamps")
        modified_dict = util.remove_output_params(
            modified_dict, "include_info_sections"
        )
        modified_dict = util.remove_output_params(modified_dict, "schedules_pings")
        return modified_dict

    def identifier(self) -> str:
        """
        Used for the "generated from ..." comment in the output.
        """
        return self.name



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
High-level interface for translating `metrics.yaml` into other formats.
"""

from pathlib import Path
import os
import shutil
import sys
import tempfile
from typing import Any, Callable, Dict, Iterable, List, Optional

from . import lint
from . import parser
from . import go_server
from . import javascript
from . import javascript_server
from . import python_server
from . import kotlin
from . import markdown
from . import metrics
from . import ruby_server
from . import rust
from . import swift
from . import util


class Outputter:
    """
    Class to define an output format.

    Each outputter in the table has the following member values:

    - output_func: the main function of the outputter, the one which
      does the actual translation.

    - clear_patterns: A list of glob patterns to clear in the directory before
      writing new results to it.
    """

    def __init__(
        self,
        output_func: Callable[[metrics.ObjectTree, Path, Dict[str, Any]], None],
        clear_patterns: Optional[List[str]] = None,
    ):
        if clear_patterns is None:
            clear_patterns = []

        self.output_func = output_func
        self.clear_patterns = clear_patterns


OUTPUTTERS = {
    "go_server": Outputter(go_server.output_go, []),
    "javascript": Outputter(javascript.output_javascript, []),
    "typescript": Outputter(javascript.output_typescript, []),
    "javascript_server": Outputter(javascript_server.output_javascript, []),
    "typescript_server": Outputter(javascript_server.output_typescript, []),
    "python_server": Outputter(python_server.output_python, []),
    "ruby_server": Outputter(ruby_server.output_ruby, []),
    "kotlin": Outputter(kotlin.output_kotlin, ["*.kt"]),
    "markdown": Outputter(markdown.output_markdown, []),
    "swift": Outputter(swift.output_swift, ["*.swift"]),
    "rust": Outputter(rust.output_rust, []),
}


def transform_metrics(objects):
    """
    Transform the object model from one that represents the YAML definitions
    to one that reflects the type specifics needed by code generators.

    e.g. This will transform a `rate` to be a `numerator` if its denominator is
    external.
    """
    counters = {}
    numerators_by_denominator: Dict[str, Any] = {}
    for category_name, category_val in objects.items():
        if category_name == "tags":
            continue
        for metric in category_val.values():
            fqmn = metric.identifier()
            if getattr(metric, "type", None) == "counter":
                counters[fqmn] = metric
            denominator_name = getattr(metric, "denominator_metric", None)
            if denominator_name:
                metric.type = "numerator"
                numerators_by_denominator.setdefault(denominator_name, [])
                numerators_by_denominator[denominator_name].append(metric)

    for denominator_name, numerators in numerators_by_denominator.items():
        if denominator_name not in counters:
            raise ValueError(
                f"No `counter` named {denominator_name} found to be used as"
                "denominator for {numerators}",
                file=sys.stderr,
            )
        counters[denominator_name].__class__ = metrics.Denominator
        counters[denominator_name].type = "denominator"
        counters[denominator_name].numerators = numerators


def translate_metrics(
    input_filepaths: Iterable[Path],
    output_dir: Path,
    translation_func: Callable[[metrics.ObjectTree, Path, Dict[str, Any]], None],
    clear_patterns: Optional[List[str]] = None,
    options: Optional[Dict[str, Any]] = None,
    parser_config: Optional[Dict[str, Any]] = None,
):
    """
    Translate the files in `input_filepaths` by running the metrics through a
    translation function and writing the results in `output_dir`.

    :param input_filepaths: list of paths to input metrics.yaml files
    :param output_dir: the path to the output directory
    :param translation_func: the function that actually performs the translation.
        It is passed the following arguments:

            - metrics_objects: The tree of metrics as pings as returned by
              `parser.parse_objects`.
            - output_dir: The path to the output directory.
            - options: A dictionary of output format-specific options.

        Examples of translation functions are in `kotlin.py` and `swift.py`.
    :param clear_patterns: a list of glob patterns of files to clear before
        generating the output files. By default, no files will be cleared (i.e.
        the directory should be left alone).
    :param options: dictionary of options. The available options are backend
        format specific. These are passed unchanged to `translation_func`.
    :param parser_config: A dictionary of options that change parsing behavior.
        See `parser.parse_metrics` for more info.
    """
    if clear_patterns is None:
        clear_patterns = []

    if options is None:
        options = {}

    if parser_config is None:
        parser_config = {}

    input_filepaths = util.ensure_list(input_filepaths)

    allow_missing_files = parser_config.get("allow_missing_files", False)
    if not input_filepaths and not allow_missing_files:
        print("❌ No metric files specified. ", end="")
        print("Use `--allow-missing-files` to not treat this as an error.")
        return 1

    if lint.glinter(input_filepaths, parser_config):
        return 1

    all_objects = parser.parse_objects(input_filepaths, parser_config)

    if util.report_validation_errors(all_objects):
        return 1

    # allow_reserved is also relevant to the translators, so copy it there
    if parser_config.get("allow_reserved"):
        options["allow_reserved"] = True

    # We don't render tags anywhere yet.
    all_objects.value.pop("tags", None)

    # Apply additional general transformations to all metrics
    transform_metrics(all_objects.value)

    # Write everything out to a temporary directory, and then move it to the
    # real directory, for transactional integrity.
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir_path = Path(tempdir)
        translation_func(all_objects.value, tempdir_path, options)

        if output_dir.is_file():
            output_dir.unlink()
        elif output_dir.is_dir() and len(clear_patterns):
            for clear_pattern in clear_patterns:
                for filepath in output_dir.glob(clear_pattern):
                    filepath.unlink()
            if len(list(output_dir.iterdir())):
                print(f"Extra contents found in '{output_dir}'.")

        # We can't use shutil.copytree alone if the directory already exists.
        # However, if it doesn't exist, make sure to create one otherwise
        # shutil.copy will fail.
        os.makedirs(str(output_dir), exist_ok=True)
        for filename in tempdir_path.glob("*"):
            shutil.copy(str(filename), str(output_dir))

    return 0


def translate(
    input_filepaths: Iterable[Path],
    output_format: str,
    output_dir: Path,
    options: Optional[Dict[str, Any]] = None,
    parser_config: Optional[Dict[str, Any]] = None,
):
    """
    Translate the files in `input_filepaths` to the given `output_format` and
    put the results in `output_dir`.

    :param input_filepaths: list of paths to input metrics.yaml files
    :param output_format: the name of the output format
    :param output_dir: the path to the output directory
    :param options: dictionary of options. The available options are backend
        format specific.
    :param parser_config: A dictionary of options that change parsing behavior.
        See `parser.parse_metrics` for more info.
    """
    if options is None:
        options = {}

    if parser_config is None:
        parser_config = {}

    format_desc = OUTPUTTERS.get(output_format, None)

    if format_desc is None:
        raise ValueError(f"Unknown output format '{output_format}'")

    return translate_metrics(
        input_filepaths,
        output_dir,
        format_desc.output_func,
        format_desc.clear_patterns,
        options,
        parser_config,
    )



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Produce coverage reports from the raw information produced by the
`GLEAN_TEST_COVERAGE` feature.
"""

import json
from .metrics import ObjectTree
from pathlib import Path
import sys
from typing import Any, Dict, List, Optional, Sequence, Set


from . import parser
from . import util


def _outputter_codecovio(metrics: ObjectTree, output_path: Path):
    """
    Output coverage in codecov.io format as defined here:

        https://docs.codecov.io/docs/codecov-custom-coverage-format

    :param metrics: The tree of metrics, already annotated with coverage by
        `_annotate_coverage`.
    :param output_path: The file to output to.
    """
    coverage: Dict[str, List] = {}
    for category in metrics.values():
        for metric in category.values():
            defined_in = metric.defined_in
            if defined_in is not None:
                path = defined_in["filepath"]
                if path not in coverage:
                    with open(path) as fd:
                        nlines = len(list(fd.readlines()))
                    lines = [None] * nlines
                    coverage[path] = lines
                file_section = coverage[path]
                file_section[int(defined_in["line"])] = getattr(metric, "covered", 0)

    with open(output_path, "w") as fd:
        json.dump({"coverage": coverage}, fd)


OUTPUTTERS = {"codecovio": _outputter_codecovio}


def _annotate_coverage(metrics, coverage_entries):
    """
    Annotate each metric with whether it is covered. Sets the attribute
    `covered` to 1 on each metric that is covered.
    """
    mapping = {}
    for category in metrics.values():
        for metric in category.values():
            mapping[metric.identifier()] = metric

    for entry in coverage_entries:
        metric_id = _coverage_entry_to_metric_id(entry)
        if metric_id in mapping:
            mapping[metric_id].covered = 1


def _coverage_entry_to_metric_id(entry: str) -> str:
    """
    Convert a coverage entry to a metric id.

    Technically, the coverage entries are rkv database keys, so are not just
    the metric identifier. This extracts the metric identifier part out.
    """
    # If getting a glean error count, report it as covering the metric the
    # error occurred in, not the `glean.error.*` metric itself.
    if entry.startswith("glean.error."):
        entry = entry.split("/")[-1]
    # If a labeled metric, strip off the label part
    return entry.split("/")[0]


def _read_coverage_entries(coverage_reports: List[Path]) -> Set[str]:
    """
    Read coverage entries from one or more files, and deduplicates them.
    """
    entries = set()

    for coverage_report in coverage_reports:
        with open(coverage_report) as fd:
            for line in fd.readlines():
                entries.add(line.strip())

    return entries


def coverage(
    coverage_reports: List[Path],
    metrics_files: Sequence[Path],
    output_format: str,
    output_file: Path,
    parser_config: Optional[Dict[str, Any]] = None,
    file=sys.stderr,
) -> int:
    """
    Commandline helper for coverage.

    :param coverage_reports: List of coverage report files, output from the
        Glean SDK when the `GLEAN_TEST_COVERAGE` environment variable is set.
    :param metrics_files: List of Path objects to load metrics from.
    :param output_format: The coverage output format to produce. Must be one of
        `OUTPUTTERS.keys()`.
    :param output_file: Path to output coverage report to.
    :param parser_config: Parser configuration object, passed to
      `parser.parse_objects`.
    :return: Non-zero if there were any errors.
    """

    if parser_config is None:
        parser_config = {}

    if output_format not in OUTPUTTERS:
        raise ValueError(f"Unknown outputter {output_format}")

    metrics_files = util.ensure_list(metrics_files)

    all_objects = parser.parse_objects(metrics_files, parser_config)

    if util.report_validation_errors(all_objects):
        return 1

    entries = _read_coverage_entries(coverage_reports)

    _annotate_coverage(all_objects.value, entries)

    OUTPUTTERS[output_format](all_objects.value, output_file)

    return 0



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from glean_parser.util import to_camel_case, remove_output_params


def test_camel_case_first_lowercase():
    assert "testMe" == to_camel_case("test_me", False)


def test_camel_case_first_uppercase():
    assert "TestMe" == to_camel_case("test_me", True)


def test_camel_case_empty_tokens():
    assert "testMe" == to_camel_case("__test____me", False)


def test_camel_case_dots_sanitized():
    assert "testMeYeah" == to_camel_case("__test..me.yeah", False)


def test_camel_case_numbers():
    assert "g33kS4n1t1z3d" == to_camel_case("g33k_s4n1t1z3d", False)


def test_camel_case_expected():
    assert "easyOne" == to_camel_case("easy_one", False)
    assert "moreInvolved1" == to_camel_case("more_involved_1", False)


def test_removing_output_params():
    d = {"name": "test dict", "defined_in": {"line": "42"}, "abc": "xyz"}

    output_removed = {"name": "test dict", "abc": "xyz"}

    test = remove_output_params(d, "defined_in")

    assert test == output_removed



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Produce skeleton Data Review Requests.
"""

from pathlib import Path
from typing import Sequence
import re


from . import parser
from . import util


def generate(
    bug: str,
    metrics_files: Sequence[Path],
) -> int:
    """
    Commandline helper for Data Review Request template generation.

    :param bug: pattern to match in metrics' bug_numbers lists.
    :param metrics_files: List of Path objects to load metrics from.
    :return: Non-zero if there were any errors.
    """

    metrics_files = util.ensure_list(metrics_files)

    # Accept any value of expires.
    parser_options = {
        "allow_reserved": True,
        "custom_is_expired": lambda expires: False,
        "custom_validate_expires": lambda expires: True,
    }
    all_objects = parser.parse_objects(metrics_files, parser_options)

    if util.report_validation_errors(all_objects):
        return 1

    # I tried [\W\Z] but it complained. So `|` it is.
    reobj = re.compile(f"\\W{bug}\\W|\\W{bug}$")
    durations = set()
    responsible_emails = set()
    filtered_metrics = list()
    for metrics in all_objects.value.values():
        for metric in metrics.values():
            if not any([len(reobj.findall(bug)) == 1 for bug in metric.bugs]):
                continue

            filtered_metrics.append(metric)

            durations.add(metric.expires)

            if metric.expires == "never":
                responsible_emails.update(metric.notification_emails)

    if len(filtered_metrics) == 0:
        print(f"I'm sorry, I couldn't find metrics matching the bug number {bug}.")
        return 1

    template = util.get_jinja2_template(
        "data_review.jinja2",
        filters=(("snake_case", util.snake_case),),
    )

    print(
        template.render(
            metrics=filtered_metrics,
            durations=durations,
            responsible_emails=responsible_emails,
        )
    )

    return 0



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Validates the contents of a Glean ping against the schema.
"""

import functools
import io
import json
from pathlib import Path
import sys

import jsonschema  # type: ignore

from . import util


ROOT_DIR = Path(__file__).parent
SCHEMAS_DIR = ROOT_DIR / "schemas"


@functools.lru_cache(maxsize=1)
def _get_ping_schema(schema_url):
    contents = util.fetch_remote_url(schema_url)
    return json.loads(contents)


def _validate_ping(ins, outs, schema_url):
    schema = _get_ping_schema(schema_url)

    resolver = util.get_null_resolver(schema)

    document = json.load(ins)

    validator_class = jsonschema.validators.validator_for(schema)
    validator = validator_class(schema, resolver=resolver)

    has_error = 0
    for error in validator.iter_errors(document):
        outs.write("=" * 76)
        outs.write("\n")
        outs.write(util.format_error("", "", util.pprint_validation_error(error)))
        outs.write("\n")
        has_error = 1

    return has_error


def validate_ping(ins, outs=None, schema_url=None):
    """
    Validates the contents of a Glean ping.

    :param ins: Input stream or file path to the ping contents to validate
    :param outs: Output stream to write errors to. (Defaults to stdout)
    :param schema_url: HTTP URL or local filesystem path to Glean ping schema.
        Defaults to the current version of the schema in
        mozilla-pipeline-schemas.
    :rtype: int 1 if any errors occurred, otherwise 0.
    """
    if schema_url is None:
        raise TypeError("Missing required argument 'schema_url'")

    if outs is None:
        outs = io.TextIOWrapper(sys.stdout.buffer, encoding="utf-8")

    if isinstance(ins, (str, bytes, Path)):
        with open(ins, "r", encoding="utf-8") as fd:
            return _validate_ping(fd, outs, schema_url=schema_url)
    else:
        return _validate_ping(ins, outs, schema_url=schema_url)



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path
import io
import json

import shutil
import subprocess


from glean_parser import translate
from glean_parser import validate_ping


ROOT = Path(__file__).parent


def test_no_event_metrics(tmp_path):
    """If no event metrics are provided parser should fail
    and no files should be generated"""

    translate.translate(
        [],
        "python_server",
        tmp_path,
    )
    assert all(False for _ in tmp_path.iterdir())


def test_logging(tmp_path):
    """Test that generated code can be used to log events in the right format."""

    glean_module_path = tmp_path / "glean"

    # generate logging code
    translate.translate(
        [
            ROOT / "data" / "server_metrics_with_event.yaml",
        ],
        "python_server",
        glean_module_path,
    )

    # copy ROOT / "test-py" / "test.py" to tmpdir
    shutil.copy(ROOT / "test-py" / "test.py", tmp_path)

    # run test script
    logged_output = subprocess.check_output(["python", "test.py"], cwd=tmp_path).decode(
        "utf-8"
    )

    logged_output = json.loads(logged_output)
    fields = logged_output["Fields"]
    payload = fields["payload"]

    # validate that ping envelope contains all the required fields
    assert "glean-server-event" == logged_output["Type"]
    assert "accounts_backend" == fields["document_namespace"]
    assert "events" == fields["document_type"]
    assert "1" == fields["document_version"]
    assert "Mozilla/5.0 ..." == fields["user_agent"]

    schema_url = (
        "https://raw.githubusercontent.com/mozilla-services/"
        "mozilla-pipeline-schemas/main/"
        "schemas/glean/glean/glean.1.schema.json"
    )

    # validate that ping payload is valid against glean schema
    input = io.StringIO(payload)
    output = io.StringIO()
    assert (
        validate_ping.validate_ping(input, output, schema_url=schema_url) == 0
    ), output.getvalue()



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Outputter to generate Javascript code for metrics.
"""

import enum
import json
from pathlib import Path
from typing import Any, Dict, Optional, Callable

from . import __version__
from . import metrics
from . import util


def javascript_datatypes_filter(value: util.JSONType) -> str:
    """
    A Jinja2 filter that renders Javascript literals.

    Based on Python's JSONEncoder, but overrides:
      - lists to use listOf
      - sets to use setOf
      - Rate objects to a CommonMetricData initializer
        (for external Denominators' Numerators lists)
    """

    class JavascriptEncoder(json.JSONEncoder):
        def iterencode(self, value):
            if isinstance(value, enum.Enum):
                yield from super().iterencode(util.camelize(value.name))
            elif isinstance(value, list):
                yield "["
                first = True
                for subvalue in value:
                    if not first:
                        yield ", "
                    yield from self.iterencode(subvalue)
                    first = False
                yield "]"
            elif isinstance(value, set):
                yield "["
                first = True
                for subvalue in sorted(list(value)):
                    if not first:
                        yield ", "
                    yield from self.iterencode(subvalue)
                    first = False
                yield "]"
            elif isinstance(value, metrics.Rate):
                yield "CommonMetricData("
                first = True
                for arg_name in util.common_metric_args:
                    if hasattr(value, arg_name):
                        if not first:
                            yield ", "
                        yield f"{util.camelize(arg_name)} = "
                        yield from self.iterencode(getattr(value, arg_name))
                        first = False
                yield ")"
            else:
                yield from super().iterencode(value)

    return "".join(JavascriptEncoder().iterencode(value))


def class_name_factory(platform: str) -> Callable[[str], str]:
    """
    Returns a function that receives an obj_type and
    returns the correct class name for that type in the current platform.
    """

    def class_name(obj_type: str) -> str:
        if obj_type == "ping":
            class_name = "PingType"
        else:
            if obj_type.startswith("labeled_"):
                obj_type = obj_type[8:]
            class_name = util.Camelize(obj_type) + "MetricType"

        if platform == "qt":
            return "Glean.Glean._private." + class_name

        return class_name

    return class_name


def extra_type_name(extra_type: str) -> str:
    """
    Returns the equivalent TypeScript type to an extra type.
    """
    if extra_type == "quantity":
        return "number"

    return extra_type


def import_path(obj_type: str) -> str:
    """
    Returns the import path of the given object inside the @mozilla/glean package.
    """
    if obj_type == "ping":
        import_path = "ping"
    else:
        if obj_type.startswith("labeled_"):
            obj_type = obj_type[8:]
        import_path = "metrics/" + obj_type

    return import_path


def args(obj_type: str) -> Dict[str, object]:
    """
    Returns the list of arguments for each object type.
    """
    if obj_type == "ping":
        return {"common": util.ping_args, "extra": []}

    return {"common": util.common_metric_args, "extra": util.extra_metric_args}


def generate_build_date(date: Optional[str]) -> str:
    """
    Generate the build Date object.
    """

    ts = util.build_date(date)

    data = [
        str(ts.year),
        # In JavaScript the first month of the year in calendars is JANUARY which is 0.
        # In Python it's 1-based
        str(ts.month - 1),
        str(ts.day),
        str(ts.hour),
        str(ts.minute),
        str(ts.second),
    ]
    components = ", ".join(data)

    # DatetimeMetricType takes a `Date` instance.
    return f"new Date({components})"  # noqa


def output(
    lang: str,
    objs: metrics.ObjectTree,
    output_dir: Path,
    options: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Given a tree of objects, output Javascript or Typescript code to `output_dir`.

    :param lang: Either "javascript" or "typescript";
    :param objects: A tree of objects (metrics and pings) as returned from
        `parser.parse_objects`.
    :param output_dir: Path to an output directory to write to.
    :param options: options dictionary, with the following optional keys:
        - `platform`: Which platform are we building for. Options are `webext` and `qt`.
                      Default is `webext`.
        - `version`: The version of the Glean.js Qt library being used.
                     This option is mandatory when targeting Qt. Note that the version
                     string must only contain the major and minor version i.e. 0.14.
        - `with_buildinfo`: If "true" a `gleanBuildInfo.(js|ts)` file is generated.
            Otherwise generation of that file is skipped. Defaults to "false".
        - `build_date`: If set to `0` a static unix epoch time will be used.
                        If set to a ISO8601 datetime string (e.g. `2022-01-03T17:30:00`)
                        it will use that date.
                        Other values will throw an error.
                        If not set it will use the current date & time.
    """

    if options is None:
        options = {}

    platform = options.get("platform", "webext")
    accepted_platforms = ["qt", "webext", "node"]
    if platform not in accepted_platforms:
        raise ValueError(
            f"Unknown platform: {platform}. Accepted platforms are: {accepted_platforms}."  # noqa
        )
    version = options.get("version")
    if platform == "qt" and version is None:
        raise ValueError(
            "'version' option is required when building for the 'qt' platform."
        )

    template = util.get_jinja2_template(
        "javascript.jinja2",
        filters=(
            ("class_name", class_name_factory(platform)),
            ("extra_type_name", extra_type_name),
            ("import_path", import_path),
            ("js", javascript_datatypes_filter),
            ("args", args),
        ),
    )

    for category_key, category_val in objs.items():
        extension = ".js" if lang == "javascript" else ".ts"
        filename = util.camelize(category_key) + extension
        filepath = output_dir / filename

        types = set(
            [
                # This takes care of the regular metric type imports
                # as well as the labeled metric subtype imports,
                # thus the removal of the `labeled_` substring.
                #
                # The actual LabeledMetricType import is conditioned after
                # the `has_labeled_metrics` boolean.
                obj.type if not obj.type.startswith("labeled_") else obj.type[8:]
                for obj in category_val.values()
            ]
        )
        has_labeled_metrics = any(
            getattr(metric, "labeled", False) for metric in category_val.values()
        )
        with filepath.open("w", encoding="utf-8") as fd:
            fd.write(
                template.render(
                    parser_version=__version__,
                    category_name=category_key,
                    objs=category_val,
                    extra_args=util.extra_args,
                    platform=platform,
                    version=version,
                    has_labeled_metrics=has_labeled_metrics,
                    types=types,
                    lang=lang,
                )
            )
            # Jinja2 squashes the final newline, so we explicitly add it
            fd.write("\n")

    with_buildinfo = options.get("with_buildinfo", "").lower() == "true"
    build_date = options.get("build_date", None)
    if with_buildinfo:
        # Write out the special "build info" file
        template = util.get_jinja2_template(
            "javascript.buildinfo.jinja2",
        )
        # This filename needs to start with "glean" so it can never
        # clash with a metric category
        filename = "gleanBuildInfo" + extension
        filepath = output_dir / filename

        with filepath.open("w", encoding="utf-8") as fd:
            fd.write(
                template.render(
                    parser_version=__version__,
                    platform=platform,
                    build_date=generate_build_date(build_date),
                )
            )
            fd.write("\n")

    if platform == "qt":
        # Explicitly create a qmldir file when building for Qt
        template = util.get_jinja2_template("qmldir.jinja2")
        filepath = output_dir / "qmldir"

        with filepath.open("w", encoding="utf-8") as fd:
            fd.write(
                template.render(
                    parser_version=__version__, categories=objs.keys(), version=version
                )
            )
            # Jinja2 squashes the final newline, so we explicitly add it
            fd.write("\n")


def output_javascript(
    objs: metrics.ObjectTree, output_dir: Path, options: Optional[Dict[str, Any]] = None
) -> None:
    """
    Given a tree of objects, output Javascript code to `output_dir`.

    :param objects: A tree of objects (metrics and pings) as returned from
        `parser.parse_objects`.
    :param output_dir: Path to an output directory to write to.
    :param options: options dictionary, with the following optional keys:

        - `namespace`: The identifier of the global variable to assign to.
                       This will only have and effect for Qt and static web sites.
                       Default is `Glean`.
        - `platform`: Which platform are we building for. Options are `webext` and `qt`.
                      Default is `webext`.
    """

    output("javascript", objs, output_dir, options)


def output_typescript(
    objs: metrics.ObjectTree, output_dir: Path, options: Optional[Dict[str, Any]] = None
) -> None:
    """
    Given a tree of objects, output Typescript code to `output_dir`.

    # Note

    The only difference between the typescript and javascript templates,
    currently is the file extension.

    :param objects: A tree of objects (metrics and pings) as returned from
        `parser.parse_objects`.
    :param output_dir: Path to an output directory to write to.
    :param options: options dictionary, with the following optional keys:

        - `namespace`: The identifier of the global variable to assign to.
                       This will only have and effect for Qt and static web sites.
                       Default is `Glean`.
        - `platform`: Which platform are we building for. Options are `webext` and `qt`.
                      Default is `webext`.
    """

    output("typescript", objs, output_dir, options)



from glean.server_events import create_events_server_event_logger


logger = create_events_server_event_logger(
    application_id="accounts_backend",
    app_display_version="0.0.1",
    channel="nightly",
)

logger.record_backend_object_update(
    user_agent="Mozilla/5.0 ...",
    ip_address="2a02:a311:803c:6300:4074:5cf2:91ac:d546",
    identifiers_fxa_account_id="test-py-project",
    object_type="some_object_type",
    object_state="some_object_state",
    linking=True,
)



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Outputter to generate server ruby code for collecting events.

This outputter is different from the rest of the outputters in that the code it
generates does not use the Glean SDK. It is meant to be used to collect events
using "events as pings" pattern in server-side environments. In these environments
SDK assumptions to measurement window and connectivity don't hold.
Generated code takes care of assembling pings with metrics, serializing to messages
conforming to Glean schema, and logging using a standard Ruby logger.
Then it's the role of the ingestion pipeline to pick the messages up and process.

Warning: this outputter supports a limited set of metrics,
see `SUPPORTED_METRIC_TYPES` below.
"""

from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional

from . import __version__, metrics, util

SUPPORTED_METRIC_TYPES = ["string", "event"]


def ping_class_name(pingName: str) -> str:
    return f"Glean{util.Camelize(pingName)}Logger"


def generate_metric_name(metric: metrics.Metric) -> str:
    return f"{metric.category}.{metric.name}"


def generate_metric_argument_name(metric: metrics.Metric) -> str:
    return f"{metric.category}_{metric.name}"


def generate_metric_argument_description(metric: metrics.Metric) -> str:
    return metric.description.replace("\n", " ").rstrip()


def event_class_name(metric: metrics.Metric) -> str:
    return f"{util.Camelize(generate_metric_argument_name(metric))}Event"


def output_ruby(
    objs: metrics.ObjectTree, output_dir: Path, options: Optional[Dict[str, Any]]
) -> None:
    """
    Given a tree of objects, output ruby code to `output_dir`.

    The output is a single file containing all the code for assembling pings with
    metrics, serializing, and submitting.

    :param objects: A tree of objects (metrics and pings) as returned from
        `parser.parse_objects`.
    :param output_dir: Path to an output directory to write to.
    """

    template = util.get_jinja2_template(
        "ruby_server.jinja2",
        filters=(
            ("ping_class_name", ping_class_name),
            ("metric_name", generate_metric_name),
            ("metric_argument_name", generate_metric_argument_name),
            ("metric_argument_description", generate_metric_argument_description),
            ("event_class_name", event_class_name),
        ),
    )

    # In this environment we don't use a concept of measurement window for collecting
    # metrics. Only "events as pings" are supported.
    # For each ping we generate code which contains all the logic for assembling it
    # with metrics, serializing, and submitting. Therefore we don't generate classes for
    # each metric as in standard outputters.
    PING_METRIC_ERROR_MSG = (
        " Server-side environment is simplified and only supports the events ping type."
        + " You should not be including pings.yaml with your parser call"
        + " or referencing any other pings in your metric configuration."
    )
    if "pings" in objs:
        print("❌ Ping definition found." + PING_METRIC_ERROR_MSG)
        return

    # Go through all metrics in objs and build a map of
    # ping->list of metric categories->list of metrics
    # for easier processing in the template.
    ping_to_metrics: Dict[str, Dict[str, List[metrics.Metric]]] = defaultdict(dict)
    for _category_key, category_val in objs.items():
        for _metric_name, metric in category_val.items():
            if isinstance(metric, metrics.Metric):
                if metric.type not in SUPPORTED_METRIC_TYPES:
                    print(
                        "❌ Ignoring unsupported metric type: "
                        + f"{metric.type}:{metric.name}."
                        + " Reach out to Glean team to add support for this"
                        + " metric type."
                    )
                    continue
                for ping in metric.send_in_pings:
                    if ping != "events":
                        (
                            print(
                                "❌ Non-events ping reference found."
                                + PING_METRIC_ERROR_MSG
                                + f"Ignoring the {ping} ping type."
                            )
                        )
                        continue
                    metrics_by_type = ping_to_metrics[ping]
                    metrics_list = metrics_by_type.setdefault(metric.type, [])
                    metrics_list.append(metric)
    if "event" not in ping_to_metrics["events"]:
        print("❌ No event metrics found...at least one event metric is required")
        return
    extension = ".rb"
    filepath = output_dir / ("server_events" + extension)
    with filepath.open("w", encoding="utf-8") as fd:
        fd.write(
            template.render(
                parser_version=__version__,
                pings=ping_to_metrics,
            )
        )



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Outputter to generate Markdown documentation for metrics.
"""

from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlsplit, parse_qs


from . import __version__
from . import metrics
from . import pings
from . import util
from collections import defaultdict


def extra_info(obj: Union[metrics.Metric, pings.Ping]) -> List[Tuple[str, str]]:
    """
    Returns a list of string to string tuples with extra information for the type
    (e.g. extra keys for events) or an empty list if nothing is available.
    """
    extra_info = []

    if isinstance(obj, metrics.Event):
        for key in obj.allowed_extra_keys:
            extra_info.append((key, obj.extra_keys[key]["description"]))

    if isinstance(obj, metrics.Labeled) and obj.ordered_labels is not None:
        for label in obj.ordered_labels:
            extra_info.append((label, None))

    if isinstance(obj, metrics.Quantity):
        extra_info.append(("unit", obj.unit))

    return extra_info


def ping_desc(
    ping_name: str, custom_pings_cache: Optional[Dict[str, pings.Ping]] = None
) -> str:
    """
    Return a text description of the ping. If a custom_pings_cache
    is available, look in there for non-reserved ping names description.
    """
    desc = ""

    if ping_name in pings.RESERVED_PING_NAMES:
        desc = (
            "This is a built-in ping that is assembled out of the "
            "box by the Glean SDK."
        )
    elif ping_name == "all-pings":
        desc = "These metrics are sent in every ping."
    elif custom_pings_cache is not None and ping_name in custom_pings_cache:
        desc = custom_pings_cache[ping_name].description

    return desc


def metrics_docs(obj_name: str) -> str:
    """
    Return a link to the documentation entry for the Glean SDK metric of the
    requested type.
    """
    # We need to fixup labeled stuff, as types are singular and docs refer
    # to them as plural.
    fixedup_name = obj_name
    if obj_name.startswith("labeled_"):
        fixedup_name += "s"

    return f"https://mozilla.github.io/glean/book/user/metrics/{fixedup_name}.html"


def ping_docs(ping_name: str) -> str:
    """
    Return a link to the documentation entry for the requested Glean SDK
    built-in ping.
    """
    if ping_name not in pings.RESERVED_PING_NAMES:
        return ""

    return f"https://mozilla.github.io/glean/book/user/pings/{ping_name}.html"


def if_empty(
    ping_name: str, custom_pings_cache: Optional[Dict[str, pings.Ping]] = None
) -> bool:
    if custom_pings_cache is not None and ping_name in custom_pings_cache:
        return custom_pings_cache[ping_name].send_if_empty
    else:
        return False


def ping_reasons(
    ping_name: str, custom_pings_cache: Dict[str, pings.Ping]
) -> Dict[str, str]:
    """
    Returns the reasons dictionary for the ping.
    """
    if ping_name == "all-pings":
        return {}
    elif ping_name in custom_pings_cache:
        return custom_pings_cache[ping_name].reasons

    return {}


def ping_data_reviews(
    ping_name: str, custom_pings_cache: Optional[Dict[str, pings.Ping]] = None
) -> Optional[List[str]]:
    if custom_pings_cache is not None and ping_name in custom_pings_cache:
        return custom_pings_cache[ping_name].data_reviews
    else:
        return None


def ping_review_title(data_url: str, index: int) -> str:
    """
    Return a title for a data review in human readable form.

    :param data_url: A url for data review.
    :param index: Position of the data review on list (e.g: 1, 2, 3...).
    """
    url_object = urlsplit(data_url)

    # Bugzilla urls like `https://bugzilla.mozilla.org/show_bug.cgi?id=1581647`
    query = url_object.query
    params = parse_qs(query)

    # GitHub urls like `https://github.com/mozilla-mobile/fenix/pull/1707`
    path = url_object.path
    short_url = path[1:].replace("/pull/", "#")

    if params and params["id"]:
        return f"Bug {params['id'][0]}"
    elif url_object.netloc == "github.com":
        return short_url

    return f"Review {index}"


def ping_bugs(
    ping_name: str, custom_pings_cache: Optional[Dict[str, pings.Ping]] = None
) -> Optional[List[str]]:
    if custom_pings_cache is not None and ping_name in custom_pings_cache:
        return custom_pings_cache[ping_name].bugs
    else:
        return None


def ping_include_client_id(
    ping_name: str, custom_pings_cache: Optional[Dict[str, pings.Ping]] = None
) -> bool:
    if custom_pings_cache is not None and ping_name in custom_pings_cache:
        return custom_pings_cache[ping_name].include_client_id
    else:
        return False


def data_sensitivity_numbers(
    data_sensitivity: Optional[List[metrics.DataSensitivity]],
) -> str:
    if data_sensitivity is None:
        return "unknown"
    else:
        return ", ".join(str(x.value) for x in data_sensitivity)


def output_markdown(
    objs: metrics.ObjectTree, output_dir: Path, options: Optional[Dict[str, Any]] = None
) -> None:
    """
    Given a tree of objects, output Markdown docs to `output_dir`.

    This produces a single `metrics.md`. The file contains a table of
    contents and a section for each ping metrics are collected for.

    :param objects: A tree of objects (metrics and pings) as returned from
        `parser.parse_objects`.
    :param output_dir: Path to an output directory to write to.
    :param options: options dictionary, with the following optional key:
        - `project_title`: The projects title.
    """
    if options is None:
        options = {}

    # Build a dictionary that associates pings with their metrics.
    #
    # {
    #  "baseline": [
    #    { ... metric data ... },
    #    ...
    #  ],
    #  "metrics": [
    #    { ... metric data ... },
    #    ...
    #  ],
    #  ...
    # }
    #
    # This also builds a dictionary of custom pings, if available.
    custom_pings_cache: Dict[str, pings.Ping] = defaultdict()
    metrics_by_pings: Dict[str, List[metrics.Metric]] = defaultdict(list)
    for _category_key, category_val in objs.items():
        for obj in category_val.values():
            # Filter out custom pings. We will need them for extracting
            # the description
            if isinstance(obj, pings.Ping):
                custom_pings_cache[obj.name] = obj
                # Pings that have `send_if_empty` set to true,
                # might not have any metrics. They need to at least have an
                # empty array of metrics to show up on the template.
                if obj.send_if_empty and not metrics_by_pings[obj.name]:
                    metrics_by_pings[obj.name] = []

            # If this is an internal Glean metric, and we don't
            # want docs for it.
            if isinstance(obj, metrics.Metric) and not obj.is_internal_metric():
                # If we get here, obj is definitely a metric we want
                # docs for.
                for ping_name in obj.send_in_pings:
                    metrics_by_pings[ping_name].append(obj)

    # Sort the metrics by their identifier, to make them show up nicely
    # in the docs and to make generated docs reproducible.
    for ping_name in metrics_by_pings:
        metrics_by_pings[ping_name] = sorted(
            metrics_by_pings[ping_name], key=lambda x: x.identifier()
        )

    project_title = options.get("project_title", "this project")
    introduction_extra = options.get("introduction_extra")

    template = util.get_jinja2_template(
        "markdown.jinja2",
        filters=(
            ("extra_info", extra_info),
            ("metrics_docs", metrics_docs),
            ("ping_desc", lambda x: ping_desc(x, custom_pings_cache)),
            ("ping_send_if_empty", lambda x: if_empty(x, custom_pings_cache)),
            ("ping_docs", ping_docs),
            ("ping_reasons", lambda x: ping_reasons(x, custom_pings_cache)),
            ("ping_data_reviews", lambda x: ping_data_reviews(x, custom_pings_cache)),
            ("ping_review_title", ping_review_title),
            ("ping_bugs", lambda x: ping_bugs(x, custom_pings_cache)),
            (
                "ping_include_client_id",
                lambda x: ping_include_client_id(x, custom_pings_cache),
            ),
            ("data_sensitivity_numbers", data_sensitivity_numbers),
        ),
    )

    filename = "metrics.md"
    filepath = output_dir / filename

    with filepath.open("w", encoding="utf-8") as fd:
        fd.write(
            template.render(
                parser_version=__version__,
                metrics_by_pings=metrics_by_pings,
                project_title=project_title,
                introduction_extra=introduction_extra,
            )
        )
        # Jinja2 squashes the final newline, so we explicitly add it
        fd.write("\n")



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from glean_parser import parser, pings

import util


def test_reserved_ping_name():
    """
    Make sure external users can't use a reserved ping name.
    """

    for ping in pings.RESERVED_PING_NAMES:
        content = {ping: {"include_client_id": True}}

        util.add_required_ping(content)
        errors = list(parser._instantiate_pings({}, {}, content, "", {}))
        assert len(errors) == 1, f"Ping '{ping}' should not be allowed"
        assert "Ping uses a reserved name" in errors[0]

        errors = list(
            parser._instantiate_pings({}, {}, content, "", {"allow_reserved": True})
        )
        assert len(errors) == 0


def test_reserved_metrics_category():
    """
    The category "pings" can't be used by metrics -- it's reserved for pings.
    """
    content = {"pings": {"metric": {"type": "string"}}}

    util.add_required(content)
    errors = list(parser.parse_objects(content))
    assert len(errors) == 1
    assert "reserved" in errors[0]


def test_camel_case_ping_name():
    content = {"camelCasePingName": {"include_client_id": True}}

    util.add_required_ping(content)
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 1
    assert "camelCasePingName" in errors[0]


def test_snake_case_ping_name():
    content = {"snake_case_ping_name": {"include_client_id": True}}

    util.add_required_ping(content)
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 1
    assert "snake_case_ping_name" in errors[0]


def test_legacy_snake_case_ping_name():
    content = {
        "bookmarks_sync": {"include_client_id": True},
        "$schema": "moz://mozilla.org/schemas/glean/pings/1-0-0",
    }

    util.add_required_ping(content)
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 0


def test_send_if_empty():
    content = {"valid-ping": {"include_client_id": True, "send_if_empty": True}}

    util.add_required_ping(content)
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 0


def test_send_if_disabled():
    content = {"disabled-ping": {"include_client_id": True, "enabled": False}}

    util.add_required_ping(content)
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 0


def test_ping_schedule():
    content = {
        "piggyback-ping": {
            "include_client_id": True,
            "metadata": {"ping_schedule": ["trigger-ping"]},
        },
        "trigger-ping": {"include_client_id": True},
    }

    util.add_required_ping(content)
    all_pings = parser.parse_objects([content])
    errors = list(all_pings)
    assert len(errors) == 0
    assert "piggyback-ping" in all_pings.value["pings"]["trigger-ping"].schedules_pings


def test_no_self_ping_schedule():
    content = {
        "my_ping": {
            "include_client_id": True,
            "metadata": {"ping_schedule": ["my_ping"]},
        }
    }

    util.add_required_ping(content)
    errors = list(parser.parse_objects([content]))
    assert "my_ping" in errors[0]



from glean_parser import parser


def test_basic_tags():
    content = {
        "$schema": "moz://mozilla.org/schemas/glean/tags/1-0-0",
        "Tag": {"description": "Normal tag"},
        "Testing :: General": {"description": "Testing / General"},
    }
    objs = parser.parse_objects([content])

    errors = list(objs)
    assert len(errors) == 0

    tags = objs.value["tags"]
    assert set(tags.keys()) == set(["Tag", "Testing :: General"])
    assert tags["Tag"].description == content["Tag"]["description"]


def test_tags_description_required():
    content = {
        "$schema": "moz://mozilla.org/schemas/glean/tags/1-0-0",
        "Tag": {},
    }
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 1
    assert "Missing required properties: description" in errors[0]


def test_tags_extra_keys_not_allowed():
    content = {
        "$schema": "moz://mozilla.org/schemas/glean/tags/1-0-0",
        "Tag": {"description": "Normal tag", "extra": "Extra stuff"},
    }
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 1
    assert "Additional properties are not allowed" in errors[0]


def test_tags_name_too_long():
    content = {
        "$schema": "moz://mozilla.org/schemas/glean/tags/1-0-0",
        "Tag" * 80: {"description": "This name is way too long"},
    }
    errors = list(parser.parse_objects([content]))
    assert len(errors) == 1
    assert "TagTagTag' is too long" in errors[0]



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path
import io
import json
import pytest
import subprocess

import glean_parser
from glean_parser import translate
from glean_parser import validate_ping

ROOT = Path(__file__).parent


def test_parser_go_server_ping_no_metrics(tmp_path, capsys):
    """Test that no files are generated if only ping definitions
    are provided without any metrics."""
    translate.translate(
        ROOT / "data" / "server_pings.yaml",
        "go_server",
        tmp_path,
    )
    assert all(False for _ in tmp_path.iterdir())


def test_parser_go_server_ping_file(tmp_path, capsys):
    """Test that no files are generated if ping definitions
    are provided."""
    translate.translate(
        [
            ROOT / "data" / "server_metrics_with_event.yaml",
            ROOT / "data" / "server_pings.yaml",
        ],
        "go_server",
        tmp_path,
    )
    assert all(False for _ in tmp_path.iterdir())


def test_parser_go_server_metrics_no_ping(tmp_path, capsys):
    """Test that no files are generated if only metric definitions
    are provided without any events metrics."""
    translate.translate(
        ROOT / "data" / "server_metrics_no_events_no_pings.yaml",
        "go_server",
        tmp_path,
    )

    captured = capsys.readouterr()
    assert all(False for _ in tmp_path.iterdir())
    assert (
        "No event metrics found...at least one event metric is required" in captured.out
    )


def test_parser_go_server_metrics_unsupported_type(tmp_path, capsys):
    """Test that no files are generated with unsupported metric types."""
    translate.translate(
        [
            ROOT / "data" / "go_server_metrics_unsupported.yaml",
        ],
        "go_server",
        tmp_path,
    )
    captured = capsys.readouterr()
    assert "Ignoring unsupported metric type" in captured.out
    unsupported_types = [
        "boolean",
        "labeled_boolean",
        "labeled_string",
        "string_list",
        "timespan",
        "uuid",
        "url",
    ]
    for t in unsupported_types:
        assert t in captured.out


def test_parser_go_server(tmp_path):
    """Test that parser works"""
    translate.translate(
        ROOT / "data" / "go_server_metrics.yaml",
        "go_server",
        tmp_path,
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["server_events.go"])

    # Make sure generated file matches expected
    with (tmp_path / "server_events.go").open("r", encoding="utf-8") as fd:
        content = fd.read()
        with (ROOT / "data" / "server_events_compare.go").open(
            "r", encoding="utf-8"
        ) as cd:
            compare_raw = cd.read()

    glean_version = f"glean_parser v{glean_parser.__version__}"
    # use replace instead of format since Go uses { }
    compare = compare_raw.replace("{current_version}", glean_version)
    assert content == compare


def run_logger(code_dir, code):
    """
    Run the Go logger and capture the output sent to STDOUT.
    """

    tmpl_code = ""
    with open(ROOT / "test-go" / "test.go.tmpl", "r") as fp:
        tmpl_code = fp.read()

    tmpl_code = tmpl_code.replace("/* CODE */", code)

    with open(code_dir / "test.go", "w") as fp:
        fp.write(tmpl_code)

    subprocess.call(["go", "mod", "init", "glean"], cwd=code_dir)
    subprocess.call(["go", "mod", "tidy"], cwd=code_dir)

    return subprocess.check_output(["go", "run", "test.go"], cwd=code_dir).decode(
        "utf-8"
    )


@pytest.mark.go_dependency
def test_run_logging(tmp_path):
    glean_module_path = tmp_path / "glean"

    translate.translate(
        [
            ROOT / "data" / "server_metrics_with_event.yaml",
        ],
        "go_server",
        glean_module_path,
    )

    code = """
    logger.RecordEventBackendObjectUpdate(
        glean.RequestInfo{
            UserAgent: "glean-test/1.0",
            IpAddress: "127.0.0.1",
        },
        glean.EventBackendObjectUpdate{
            IdentifiersFxaAccountId: "accountId",
            ObjectType:              "type",
            ObjectState:             "state",
        },
    )
    """

    logged_output = run_logger(tmp_path, code)
    logged_output = json.loads(logged_output)
    fields = logged_output["Fields"]
    payload = fields["payload"]

    assert "glean-server-event" == logged_output["Type"]
    assert "glean.test" == fields["document_namespace"]
    assert "events" == fields["document_type"]
    assert "1" == fields["document_version"]
    assert "glean-test/1.0" == fields["user_agent"]

    schema_url = (
        "https://raw.githubusercontent.com/mozilla-services/"
        "mozilla-pipeline-schemas/main/"
        "schemas/glean/glean/glean.1.schema.json"
    )

    input = io.StringIO(payload)
    output = io.StringIO()
    assert (
        validate_ping.validate_ping(input, output, schema_url=schema_url) == 0
    ), output.getvalue()



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path
import shutil
import subprocess
import re

from glean_parser import swift
from glean_parser import metrics
from glean_parser import pings
from glean_parser import translate
from glean_parser.util import DictWrapper


ROOT = Path(__file__).parent


def run_linters(files):
    # Syntax check on the generated files.
    # Only run this test if swiftc is on the path.
    if shutil.which("swiftc"):
        for filepath in files:
            subprocess.check_call(["swiftc", "-parse", filepath])

    # Lint check on the generated files.
    # Only run this test if swiftlint is on the path.
    if shutil.which("swiftlint"):
        for filepath in files:
            subprocess.check_call(["swiftlint", "lint", filepath])


def test_parser(tmp_path):
    """Test translating metrics to Swift files."""
    translate.translate(
        ROOT / "data" / "core.yaml", "swift", tmp_path, {}, {"allow_reserved": True}
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    # Make sure descriptions made it in
    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()

        assert "True if the user has set Firefox as the default browser." in content
        assert "جمع 搜集" in content
        assert 'category: ""' in content
        assert "class GleanBuild" in content
        assert "BuildInfo(buildDate:" in content

    run_linters(tmp_path.glob("*.swift"))


def test_parser_no_build_info(tmp_path):
    """Test translating metrics to Swift files without build info."""
    translate.translate(
        ROOT / "data" / "core.yaml",
        "swift",
        tmp_path,
        {"with_buildinfo": "false"},
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    # Make sure descriptions made it in
    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()

        assert "class GleanBuild" not in content

    run_linters(tmp_path.glob("*.swift"))


def test_parser_custom_build_date(tmp_path):
    """Test translating metrics to Swift files without build info."""
    translate.translate(
        ROOT / "data" / "core.yaml",
        "swift",
        tmp_path,
        {"build_date": "2020-01-01T17:30:00"},
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    # Make sure descriptions made it in
    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()

        assert "class GleanBuild" in content
        assert "BuildInfo(buildDate:" in content
        assert "year: 2020, month: 1, day: 1" in content

    run_linters(tmp_path.glob("*.swift"))


def test_parser_all_metrics(tmp_path):
    """Test translating ALL metric types to Swift files."""
    translate.translate(
        ROOT / "data" / "all_metrics.yaml",
        "swift",
        tmp_path,
        {"namespace": "Foo"},
        {"allow_reserved": False},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    run_linters(tmp_path.glob("*.swift"))


def test_ping_parser(tmp_path):
    """Test translating pings to Swift files."""
    translate.translate(
        ROOT / "data" / "pings.yaml",
        "swift",
        tmp_path,
        {"namespace": "Foo"},
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    # Make sure descriptions made it in
    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()
        assert "This is a custom ping" in content
        # Make sure the namespace option is in effect
        assert "extension Foo" in content
        assert "customPing = Ping<NoReasonCodes>" in content
        assert (
            "customPingMightBeEmpty = Ping<CustomPingMightBeEmptyReasonCodes>"
            in content
        )

    run_linters(tmp_path.glob("*.swift"))


def test_swift_generator():
    kdf = swift.swift_datatypes_filter

    assert kdf("\n") == r'"\n"'
    assert kdf([42, "\n"]) == r'[42, "\n"]'
    assert (
        kdf(DictWrapper([("key", "value"), ("key2", "value2")]))
        == r'["key": "value", "key2": "value2"]'
    )
    assert kdf(metrics.Lifetime.ping) == ".ping"


def test_metric_type_name():
    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["42"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
        extra_keys={"my_extra": {"description": "an extra", "type": "string"}},
    )

    assert swift.type_name(event) == "EventMetricType<MetricExtra>"

    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["42"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
    )

    assert swift.type_name(event) == "EventMetricType<NoExtras>"

    boolean = metrics.Boolean(
        type="boolean",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
    )
    assert swift.type_name(boolean) == "BooleanMetricType"

    ping = pings.Ping(
        name="custom",
        description="description...",
        include_client_id=True,
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@nowhere.com"],
    )
    assert swift.type_name(ping) == "Ping<NoReasonCodes>"

    ping = pings.Ping(
        name="custom",
        description="description...",
        include_client_id=True,
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@nowhere.com"],
        reasons={"foo": "foolicious", "bar": "barlicious"},
    )
    assert swift.type_name(ping) == "Ping<CustomReasonCodes>"


def test_order_of_fields(tmp_path):
    """Test that translating metrics to Swift files keeps a stable order of fields."""
    translate.translate(
        ROOT / "data" / "core.yaml", "swift", tmp_path, {}, {"allow_reserved": True}
    )

    # Make sure descriptions made it in
    fd = (tmp_path / "Metrics.swift").open("r", encoding="utf-8")
    content = fd.read()
    fd.close()

    lines = content.splitlines()
    first_metric_fields = []
    found_metric = False

    # Get the fields of the first metric
    # Checking only one metric should be good enough for now
    for line in lines:
        if found_metric:
            if re.search("\\)$", line):
                break

            # Collect only the fields
            field = line.strip().split(":")[0]
            first_metric_fields.append(field)
        elif re.search("CommonMetricData", line):
            found_metric = True

    expected_fields = ["category", "name", "sendInPings", "lifetime", "disabled"]

    # We only check the limited list of always available fields.
    size = len(expected_fields)
    assert expected_fields == first_metric_fields[:size]


def test_no_import_glean(tmp_path):
    translate.translate(
        ROOT / "data" / "core.yaml", "swift", tmp_path, {}, {"allow_reserved": True}
    )

    # Make sure descriptions made it in
    fd = (tmp_path / "Metrics.swift").open("r", encoding="utf-8")
    content = fd.read()
    fd.close()

    assert "import Glean" not in content


def test_import_glean(tmp_path):
    translate.translate(ROOT / "data" / "smaller.yaml", "swift", tmp_path, {}, {})

    # Make sure descriptions made it in
    fd = (tmp_path / "Metrics.swift").open("r", encoding="utf-8")
    content = fd.read()
    fd.close()

    assert "import Glean" in content


def test_event_extra_keys_in_correct_order(tmp_path):
    """
    Assert that the extra keys appear in the parameter and the enumeration in
    the same order.

    https://bugzilla.mozilla.org/show_bug.cgi?id=1648768
    """
    translate.translate(
        ROOT / "data" / "event_key_ordering.yaml",
        "swift",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        assert (
            "struct ExampleExtra: EventExtras "
            "{ var and1withextracasing: Bool? var alice: String? var bob: String? var charlie: String?" in content
        )
        assert ', ["And1WithExtraCasing", "alice", "bob", "charlie"]' in content


def test_event_extra_keys_with_types(tmp_path):
    """
    Assert that the extra keys with types appear with their corresponding types.
    """
    translate.translate(
        ROOT / "data" / "events_with_types.yaml",
        "swift",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        assert (
            "struct PreferenceToggledExtra: EventExtras "
            "{ var enabled: Bool? var preference: String? "
            "var swapped: Int32?" in content
        )
        assert ', ["enabled", "preference", "swapped"]' in content


def test_reasons(tmp_path):
    """
    Assert that we generate the reason codes as a plain enum.

    https://bugzilla.mozilla.org/show_bug.cgi?id=1811888
    """
    translate.translate(
        ROOT / "data" / "pings.yaml",
        "swift",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())

    expected = "enum CustomPingMightBeEmptyReasonCodes: Int, ReasonCodes { case serious = 0 case silly = 1 public func index() -> Int { return self.rawValue } }"  # noqa
    assert expected in content

    expected = "let customPing = Ping<NoReasonCodes>("
    assert expected in content

    expected = "let customPingMightBeEmpty = Ping<CustomPingMightBeEmptyReasonCodes>("
    assert expected in content


def test_object_metric(tmp_path):
    """
    Assert that an object metric is created.
    """
    translate.translate(
        ROOT / "data" / "object.yaml",
        "swift",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["Metrics.swift"])

    with (tmp_path / "Metrics.swift").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())

        assert "typealias ArrayInArrayObjectItemItem = Bool" in content
        assert "typealias NumberArrayObjectItem = Int64" in content

        assert "ObjectMetricType<ThreadsObject>" in content
        assert "typealias ThreadsObject = [ThreadsObjectItem]" in content
        assert "struct ThreadsObjectItem: Codable, Equatable, ObjectSerialize {" in content
        assert (
            "var frames: ThreadsObjectItemFrames"
            in content
        )

        assert "struct ThreadsObjectItemFramesItem: Codable, Equatable, ObjectSerialize {" in content
        assert "var moduleIndex: Int64?" in content
        assert "var ip: String?" in content
        assert "var trust: String?" in content



# -*- coding: utf-8 -*-

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Outputter to generate Rust code for metrics.
"""

import enum
import json
from pathlib import Path
from typing import Any, Dict, Optional, Union

from . import __version__
from . import metrics
from . import pings
from . import tags
from . import util


def rust_datatypes_filter(value):
    """
    A Jinja2 filter that renders Rust literals.

    Based on Python's JSONEncoder, but overrides:
      - dicts and sets to raise an error
      - sets to vec![] (used in labels)
      - enums to become Class::Value
      - lists to vec![] (used in send_in_pings)
      - null to None
      - strings to "value".into()
      - Rate objects to a CommonMetricData initializer
        (for external Denominators' Numerators lists)
    """

    class RustEncoder(json.JSONEncoder):
        def iterencode(self, value):
            if isinstance(value, dict):
                raise ValueError("RustEncoder doesn't know dicts {}".format(str(value)))
            elif isinstance(value, enum.Enum):
                yield (value.__class__.__name__ + "::" + util.Camelize(value.name))
            elif isinstance(value, set):
                yield "vec!["
                first = True
                for subvalue in sorted(list(value)):
                    if not first:
                        yield ", "
                    yield from self.iterencode(subvalue)
                    first = False
                yield "]"
            elif isinstance(value, list):
                yield "vec!["
                first = True
                for subvalue in list(value):
                    if not first:
                        yield ", "
                    yield from self.iterencode(subvalue)
                    first = False
                yield "]"
            elif value is None:
                yield "None"
            # `CowStr` is a `str`, so needs to be before next case
            elif isinstance(value, metrics.CowString):
                yield f'::std::borrow::Cow::from("{value.inner}")'
            elif isinstance(value, str):
                yield f"{json.dumps(value)}.into()"
            elif isinstance(value, metrics.Rate):
                yield "CommonMetricData("
                first = True
                for arg_name in util.common_metric_args:
                    if hasattr(value, arg_name):
                        if not first:
                            yield ", "
                        yield f"{util.camelize(arg_name)} = "
                        yield from self.iterencode(getattr(value, arg_name))
                        first = False
                yield ")"
            else:
                yield from super().iterencode(value)

    return "".join(RustEncoder().iterencode(value))


def ctor(obj):
    """
    Returns the scope and name of the constructor to use for a metric object.
    Necessary because LabeledMetric<T> is constructed using LabeledMetric::new
    not LabeledMetric<T>::new
    """
    if getattr(obj, "labeled", False):
        return "LabeledMetric::new"
    return class_name(obj.type) + "::new"


def type_name(obj):
    """
    Returns the Rust type to use for a given metric or ping object.
    """

    if getattr(obj, "labeled", False):
        return "LabeledMetric<{}>".format(class_name(obj.type))
    generate_enums = getattr(obj, "_generate_enums", [])  # Extra Keys? Reasons?
    if len(generate_enums):
        generic = None
        for name, suffix in generate_enums:
            if len(getattr(obj, name)):
                generic = util.Camelize(obj.name) + suffix
            else:
                if isinstance(obj, metrics.Event):
                    generic = "NoExtraKeys"
                else:
                    generic = "No" + suffix

        return "{}<{}>".format(class_name(obj.type), generic)

    generate_structure = getattr(obj, "_generate_structure", [])
    if len(generate_structure):
        generic = util.Camelize(obj.name) + "Object"
        return "{}<{}>".format(class_name(obj.type), generic)

    return class_name(obj.type)


def extra_type_name(typ: str) -> str:
    """
    Returns the corresponding Rust type for event's extra key types.
    """

    if typ == "boolean":
        return "bool"
    elif typ == "string":
        return "String"
    elif typ == "quantity":
        return "u32"
    else:
        return "UNSUPPORTED"


def structure_type_name(typ: str) -> str:
    """
    Returns the corresponding Rust type for structure items.
    """

    if typ == "boolean":
        return "bool"
    elif typ == "string":
        return "String"
    elif typ == "number":
        return "i64"
    else:
        return "UNSUPPORTED"


def class_name(obj_type):
    """
    Returns the Rust class name for a given metric or ping type.
    """
    if obj_type == "ping":
        return "Ping"
    if obj_type.startswith("labeled_"):
        obj_type = obj_type[8:]
    return util.Camelize(obj_type) + "Metric"


def extra_keys(allowed_extra_keys):
    """
    Returns the &'static [&'static str] ALLOWED_EXTRA_KEYS for impl ExtraKeys
    """
    return "&[" + ", ".join([f'"{key}"' for key in allowed_extra_keys]) + "]"


class Category:
    """
    Data struct holding information about a metric to be used in the template.
    """

    def __init__(
        self,
        name: str,
        objs: Dict[str, Union[metrics.Metric, pings.Ping, tags.Tag]],
        contains_pings: bool,
    ):
        self.name = name
        self.objs = objs
        self.contains_pings = contains_pings


def output_rust(
    objs: metrics.ObjectTree, output_dir: Path, options: Optional[Dict[str, Any]] = None
) -> None:
    """
    Given a tree of objects, output Rust code to `output_dir`.

    :param objs: A tree of objects (metrics and pings) as returned from
        `parser.parse_objects`.
    :param output_dir: Path to an output directory to write to.
    :param options: options dictionary, not currently used for Rust
    """

    if options is None:
        options = {}

    template = util.get_jinja2_template(
        "rust.jinja2",
        filters=(
            ("rust", rust_datatypes_filter),
            ("snake_case", util.snake_case),
            ("camelize", util.camelize),
            ("type_name", type_name),
            ("extra_type_name", extra_type_name),
            ("structure_type_name", structure_type_name),
            ("ctor", ctor),
            ("extra_keys", extra_keys),
        ),
    )

    filename = "glean_metrics.rs"
    filepath = output_dir / filename
    categories = []

    for category_key, category_val in objs.items():
        contains_pings = any(
            isinstance(obj, pings.Ping) for obj in category_val.values()
        )

        cat = Category(category_key, category_val, contains_pings)
        categories.append(cat)

    with filepath.open("w", encoding="utf-8") as fd:
        fd.write(
            template.render(
                parser_version=__version__,
                categories=categories,
                extra_metric_args=util.extra_metric_args,
                common_metric_args=util.common_metric_args,
            )
        )



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from pathlib import Path
import shutil
import subprocess
import re

from glean_parser import rust
from glean_parser import metrics
from glean_parser import pings
from glean_parser import translate


ROOT = Path(__file__).parent


def run_linters(files):
    # Syntax check on the generated files.
    # Only run this test if cargo is on the path.
    if shutil.which("rustfmt"):
        for filepath in files:
            subprocess.check_call(
                [
                    "rustfmt",
                    "--check",
                    filepath,
                ]
            )
    if shutil.which("cargo"):
        for filepath in files:
            subprocess.check_call(
                [
                    "cargo",
                    "clippy",
                    "--all",
                    "--",
                    "-D",
                    "warnings",
                    filepath,
                ]
            )


def test_parser(tmp_path):
    """Test translating metrics to Rust files."""
    translate.translate(
        ROOT / "data" / "core.yaml", "rust", tmp_path, {}, {"allow_reserved": True}
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["glean_metrics.rs"])

    # Make sure descriptions made it in
    with (tmp_path / "glean_metrics.rs").open("r", encoding="utf-8") as fd:
        content = fd.read()

        assert "True if the user has set Firefox as the default browser." in content
        assert "جمع 搜集" in content
        assert 'category: "telemetry"' in content

    # We don't have a cargo.toml, not sure what to do here aside from creating a fake
    # one for the purpose of running cargo fmt and cargo clippy
    # run_linters(tmp_path.glob("*.rs"))


def test_ping_parser(tmp_path):
    """Test translating pings to Rust files."""
    translate.translate(
        ROOT / "data" / "pings.yaml",
        "rust",
        tmp_path,
        {"namespace": "Foo"},
        {"allow_reserved": True},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["glean_metrics.rs"])

    # Make sure descriptions made it in
    with (tmp_path / "glean_metrics.rs").open("r", encoding="utf-8") as fd:
        content = fd.read()

        assert "This is a custom ping" in content
        assert (
            "custom_ping: ::glean::private::__export::Lazy<::glean::private::"
            + "PingType> =\n    ::glean::private::__export::Lazy::new"
            in content
        )
        assert (
            "custom_ping_might_be_empty: ::glean::private::__export::Lazy<"
            + "::glean::private::PingType> =\n    ::glean::private::__export::Lazy::new"
            in content
        )

    # TODO we need a cargo.toml to run `cargo fmt` and `cargo clippy`
    # and I'm not quite sure how to do that in a non-Rust project for
    # the purpose of testing
    run_linters(tmp_path.glob("*.rs"))


def test_rust_generator():
    kdf = rust.rust_datatypes_filter

    # The Rust datatypes filter encodes strings using JSON-escaping
    assert kdf("\n") == '"\\n".into()'
    assert kdf([42, "\n"]) == 'vec![42, "\\n".into()]'
    assert kdf(metrics.Lifetime.ping) == "Lifetime::Ping"


def test_metric_type_name():
    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["42"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
        extra_keys={"my_extra": {"description": "an extra", "type": "string"}},
    )

    assert rust.type_name(event) == "EventMetric<MetricExtra>"

    event = metrics.Event(
        type="event",
        category="category",
        name="metric",
        bugs=["42"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
    )

    assert rust.type_name(event) == "EventMetric<NoExtraKeys>"

    boolean = metrics.Boolean(
        type="boolean",
        category="category",
        name="metric",
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@example.com"],
        description="description...",
        expires="never",
    )
    assert rust.type_name(boolean) == "BooleanMetric"

    ping = pings.Ping(
        name="custom",
        description="description...",
        include_client_id=True,
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@nowhere.com"],
    )
    assert rust.type_name(ping) == "Ping<NoReasonCodes>"

    ping = pings.Ping(
        name="custom",
        description="description...",
        include_client_id=True,
        bugs=["http://bugzilla.mozilla.com/12345"],
        notification_emails=["nobody@nowhere.com"],
        reasons={"foo": "foolicious", "bar": "barlicious"},
    )
    assert rust.type_name(ping) == "Ping<CustomReasonCodes>"


def test_order_of_fields(tmp_path):
    """Test that translating metrics to Rust files keeps a stable order of fields."""
    translate.translate(
        ROOT / "data" / "core.yaml", "rust", tmp_path, {}, {"allow_reserved": True}
    )

    # Make sure descriptions made it in
    fd = (tmp_path / "glean_metrics.rs").open("r", encoding="utf-8")
    content = fd.read()
    fd.close()

    lines = content.splitlines()
    first_metric_fields = []
    found_metric = False

    # Get the fields of the first metric
    # Checking only one metric should be good enough for now
    for line in lines:
        if found_metric:
            if re.search("..Default::default()$", line):
                break

            # Collect only the fields
            field = line.strip().split(":")[0]
            first_metric_fields.append(field)
        elif re.search("CommonMetricData {", line):
            found_metric = True

    expected_fields = ["category", "name", "send_in_pings", "lifetime", "disabled"]

    # We only check the limited list of always available fields.
    size = len(expected_fields)
    assert expected_fields == first_metric_fields[:size]


def test_event_extra_keys_in_correct_order(tmp_path):
    """
    Assert that the extra keys appear in the parameter and the enumeration in
    the same order.

    https://bugzilla.mozilla.org/show_bug.cgi?id=1648768
    """
    translate.translate(
        ROOT / "data" / "events_with_types.yaml",
        "rust",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["glean_metrics.rs"])

    with (tmp_path / "glean_metrics.rs").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        assert (
            "pub struct PreferenceToggledExtra { "
            "pub enabled: Option<bool>, pub preference: "
            "Option<String>, pub swapped: Option<u32>, }" in content
        )
        assert (
            "const ALLOWED_KEYS: &'static [&'static str] = "
            '&["enabled", "preference", "swapped"];' in content
        )


def test_event_extra_keys_with_types(tmp_path):
    """
    Assert that the extra keys with types appear with their corresponding types.
    """
    translate.translate(
        ROOT / "data" / "events_with_types.yaml",
        "rust",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["glean_metrics.rs"])

    with (tmp_path / "glean_metrics.rs").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())
        assert (
            "impl ExtraKeys for PreferenceToggledExtra { "
            "const ALLOWED_KEYS: &'static [&'static str] = "
            '&["enabled", "preference", "swapped"];' in content
        )
        assert (
            "const ALLOWED_KEYS: &'static [&'static str] = "
            '&["enabled", "preference", "swapped"];' in content
        )


def test_object_metric(tmp_path):
    """
    Assert that an object metric is created.
    """
    translate.translate(
        ROOT / "data" / "object.yaml",
        "rust",
        tmp_path,
        {"namespace": "Foo"},
    )

    assert set(x.name for x in tmp_path.iterdir()) == set(["glean_metrics.rs"])

    with (tmp_path / "glean_metrics.rs").open("r", encoding="utf-8") as fd:
        content = fd.read()
        content = " ".join(content.split())

        assert "ObjectMetric<ThreadsObject>" in content
        assert "pub struct ThreadsObjectItem { " in content
        assert "frames: ThreadsObjectItemFrames, }" in content
        assert (
            "pub type ThreadsObjectItemFrames = "
            "Vec<ThreadsObjectItemFramesItem>;" in content
        )

        assert "pub struct ThreadsObjectItemFramesItem { " in content
        assert "module_index: Option<i64>, " in content
        assert "ip: Option<String>, " in content
        assert "trust: Option<String>, " in content
        assert "}" in content
